# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analog Information utility.

This module prints relevant information about the model and its analog
execution.

"""

from functools import reduce
import operator
from typing import Optional, Any, List

# Imports from PyTorch.
from torch import zeros
from torch.nn import Module

from aihwkit.nn.modules.base import AnalogModuleBase, RPUConfigAlias
from aihwkit.nn.modules.conv_mapped import _AnalogConvNdMapped
from aihwkit.nn.modules.linear_mapped import AnalogLinearMapped
from aihwkit.nn.modules.conv import _AnalogConvNd
from aihwkit.nn import AnalogLinear
from aihwkit.simulator.tiles import BaseTile

COLUMN_DEFINITIONS = ["Layer Information", "Tile Information"]

COLUMN_NAMES = {
    "name": (0, "Layer Name"),
    "isanalog": (0, "Is Analog"),
    "input_size": (0, "In Shape"),
    "output_size": (0, "Out Shape"),
    "kernel_size": (0, "Kernel Shape"),
    "num_tiles": (0, "# of Tiles"),
    # "macs": "Multi_Adds",
    "log_shape": (1, "Log. tile shape"),
    "phy_shape": (1, "Phys. tile shape"),
    "utilization": (1, "utilization (%)"),
    "reuse_factor": (0, "Reuse Factor")
}

FORMATTING_WIDTH = 200
COLUMN_WIDTH = 10
FLOAT_FORMAT = "{0:.2f}"


class TileInfo:
    """Class for storing tile statistics and information."""

    log_in_size: Any
    log_out_size: Any
    phy_in_size: Any
    phy_out_size: Any
    utilization: float

    def __init__(self, tile: BaseTile, is_mapped: bool):
        self.log_in_size = tile.in_size
        self.log_out_size = tile.out_size
        self.phy_in_size = tile.rpu_config.mapping.max_input_size
        self.phy_out_size = tile.rpu_config.mapping.max_output_size
        self.is_mapped = is_mapped
        max_space = (self.phy_in_size*self.phy_out_size)
        log_space = (self.log_in_size * self.log_out_size)
        self.utilization = log_space * 100 / max_space if is_mapped else 100

    def tile_summary_dict(self) -> dict:
        """Return a dictionary with the tile info."""
        phys_shape = 'N/A' if not self.is_mapped else (self.phy_out_size, self.phy_in_size)
        return {"log_shape": str((self.log_out_size, self.log_in_size)),
                "phys_shape": str(phys_shape),
                "utilization": self.utilization
                }

    def __repr__(self) -> str:
        """Print Tile's information."""
        tile_info = self.tile_summary_dict().values()
        return "{:<20}{:<20}{:<20}\n".format(*(tile_info))


class LayerInfo:
    """Class for storing layer statistics and information."""
    # pylint: disable=too-many-instance-attributes
    module: Module
    name: str
    isanalog: bool
    num_tiles: int
    tiles_info: List[TileInfo]
    input_size: Any
    output_size: Any
    kernel_size: Any
    reuse_factor: int

    def __init__(self,
                 module: Module,
                 rpu_config: Optional[RPUConfigAlias] = None,
                 input_size: Any = None,
                 output_size: Any = None):
        self.module = module
        self.name = self.module.__class__.__name__
        self.isanalog = isinstance(self.module, AnalogModuleBase)
        self.num_tiles = 0 if not self.isanalog else self.module._analog_tile_counter
        self.tiles_info = self.set_tiles_info()
        self.kernel_size = None
        self.reuse_factor = 0
        self.input_size, self.output_size = input_size, output_size
        self.rpu_config = rpu_config
        self.set_kernel_size()
        self.calculate_reuse_factor()

    def __set_reuse_factor(self, reuse_factor: int) -> None:
        """Private method to set layer's reuse factor attribute."""
        self.reuse_factor = reuse_factor

    def set_tiles_info(self) -> List[TileInfo]:
        """Create TileInfo objects for each tile of the layer."""
        tiles_info = []
        is_mapped = isinstance(self.module, AnalogLinearMapped)
        is_mapped = is_mapped or isinstance(self.module, _AnalogConvNdMapped)
        if isinstance(self.module, AnalogModuleBase):
            for tile in self.module.analog_tiles():
                tiles_info.append(TileInfo(tile, is_mapped))
        return tiles_info

    def set_kernel_size(self) -> None:
        """Set kernel size attribute."""
        if hasattr(self.module, 'kernel_size'):
            self.kernel_size = self.module.kernel_size

    def calculate_reuse_factor(self) -> None:
        """Compute the reuse factor.

        The reuse factor is the number of vector matrix multiplication
        a layer computes.

        """
        if isinstance(self.module, (AnalogLinear, AnalogLinearMapped)):
            ruf = reduce(operator.mul, (self.input_size), 1) // int(self.input_size[-1])
            self.__set_reuse_factor(ruf)
        elif isinstance(self.module, (_AnalogConvNd, _AnalogConvNdMapped)):
            ruf = reduce(operator.mul, (self.output_size), 1) // self.output_size[1]
            self.__set_reuse_factor(ruf)

    def tiles_summary_dict(self) -> dict:
        """Return a dictionary with all tiles information."""
        tiles_summary = {}
        for tile in self.tiles_info:
            tiles_summary.update(tile.tile_summary_dict())
        return tiles_summary

    def layer_summary_dict(self) -> dict:
        """Return a dictionary with all layer's information."""

        analog = "analog" if self.isanalog else "digital"
        return {"name": self.name,
                "isanalog": analog,
                "input_size": str(self.input_size) if self.input_size is not None else "-",
                "output_size": str(self.output_size) if self.output_size is not None else "-",
                "kernel_size": str(self.kernel_size) if self.kernel_size is not None else "-",
                "num_tiles": self.num_tiles,
                "reuse_factor": str(self.reuse_factor) if self.reuse_factor is not None else "-",
                "log_shape": "-",
                "phy_shape": "-",
                "utilization": "-"}

    def __repr__(self) -> str:
        """Print layer's information in the summary table."""
        stats = self.layer_summary_dict().values()
        result = (("{:<20}"*len(stats)).format(*stats))
        result += "\n"
        for tile in self.tiles_info:
            tile_info = tile.tile_summary_dict()
            tile_info["utilization"] = FLOAT_FORMAT.format(tile_info["utilization"])
            result += (" "*20*(len(stats)-3))
            result += ("{:<20}{:<20}{:<20}\n".format(*(tile_info.values())))
        return result


class AnalogInfo:
    """Class for computing and storing results of the analog summary."""

    def __init__(self,
                 model: Module,
                 input_size: Any = None,
                 rpu_config: Optional[RPUConfigAlias] = None):

        self.model = model
        self.input_size = input_size
        self.rpu_config = rpu_config
        self.layer_summary = self.create_layer_summary()
        self.total_tile_number = self.calculate_num_tiles()
        self.total_nb_analog = self.calculate_num_analog()

    def register_hooks_recursively(self, module: Module, hook: Any) -> None:
        """Hooks the function into all layers with no children."""
        if not list(module.children()):
            module.register_forward_hook(hook)
        else:
            for _, layer in module.named_children():
                self.register_hooks_recursively(layer, hook)

    def create_layer_summary(self) -> List[LayerInfo]:
        """Create the layer summary list.

        This list contains LayerInfo elements that corresponds to each
        layer of the model.

        """
        layer_summary = []

        def get_size_hook(_: Module, _input: Any, _output: Any) -> None:
            nonlocal layer_summary
            input_size = list(_input[0].size())
            output_size = list(_output.size())
            layer_summary.append(LayerInfo(_, self.rpu_config, input_size, output_size))

        self.register_hooks_recursively(self.model, get_size_hook)
        device = next(self.model.parameters()).device
        dummy_var = zeros(self.input_size).to(device)
        self.model(dummy_var)
        return layer_summary

    def calculate_num_tiles(self) -> int:
        """Calculate the total number of tiles needed by the model."""
        total_tile_number = 0
        for x in self.layer_summary:
            total_tile_number += x.num_tiles
        return total_tile_number

    def calculate_num_analog(self) -> int:
        """Calculate the number of analog layers."""
        total_nb_analog = 0
        for x in self.layer_summary:
            total_nb_analog += 1 if x.isanalog else 0
        return total_nb_analog

    def __repr__(self) -> str:
        """Print summary results."""
        divider = "=" * FORMATTING_WIDTH + "\n"
        name = "Model Name: " + self.model.__class__.__name__ + "\n"
        result = divider + name + divider
        result += "Per-layer Information\n" + divider

        # Add header
        header = [*COLUMN_NAMES.values()]
        for i, category in enumerate(COLUMN_DEFINITIONS):
            header_i = [v for x, v in header if x == i]
            trim_length = (COLUMN_WIDTH * len(header_i) - len(category))
            result += category + " " * trim_length
            if i == len(COLUMN_DEFINITIONS)-1:
                break
            result += '| '
        result += "\n"+divider
        for i, category in enumerate(COLUMN_DEFINITIONS):
            header_i = [v for x, v in header if x == i]
            result += (("{:<20}"*len(header_i)).format(*header_i))
        result += "\n"

        for x in self.layer_summary:
            result += repr(x)

        result += divider
        result += "General Information\n" + divider
        result += "Total number of tiles: " + str(self.total_tile_number) + "\n"
        result += "Total number of analog layers: " + str(self.total_nb_analog) + "\n"
        return result


def analog_summary(model: Module,
                   input_size: Optional[Any] = None,
                   rpu_config: Optional[RPUConfigAlias] = None) -> AnalogInfo:
    """Summarize the given PyTorch model.

    Summarized information includes:

        1) Layer names,
        2) input/output shapes,
        3) kernel shape,
        4) # of digital parameters,
        5) # of analog parameters,
        6) # of analog tiles
        7) reuse factor

    Args:
        model: PyTorch model to run on the analog platform.

        input_size: required to run a forward pass of the model.

        rpu_config: resistive processing unit configuration.

    Returns:
        AnalogInfo Object.
    """
    results = AnalogInfo(model, input_size, rpu_config)
    print(results)
    return results



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: onnx_common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11onnx_common.proto\x12\x05\x61ihwx\"\x91\x03\n\x0e\x41ttributeProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x31\n\x04type\x18\x14 \x01(\x0e\x32#.aihwx.AttributeProto.AttributeType\x12\t\n\x01\x66\x18\x02 \x01(\x02\x12\t\n\x01i\x18\x03 \x01(\x03\x12\t\n\x01s\x18\x04 \x01(\x0c\x12\x1d\n\x01t\x18\x05 \x01(\x0b\x32\x12.aihwx.TensorProto\x12\t\n\x01\x62\x18\x65 \x01(\x08\x12\x0e\n\x06\x66loats\x18\x07 \x03(\x02\x12\x0c\n\x04ints\x18\x08 \x03(\x03\x12\x0f\n\x07strings\x18\t \x03(\x0c\x12#\n\x07tensors\x18\n \x03(\x0b\x32\x12.aihwx.TensorProto\x12\r\n\x05\x62ools\x18\x66 \x03(\x08\"\x8f\x01\n\rAttributeType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\x07\n\x03INT\x10\x02\x12\n\n\x06STRING\x10\x03\x12\n\n\x06TENSOR\x10\x04\x12\x08\n\x04\x42OOL\x10\x65\x12\n\n\x06\x46LOATS\x10\x06\x12\x08\n\x04INTS\x10\x07\x12\x0b\n\x07STRINGS\x10\x08\x12\x0b\n\x07TENSORS\x10\t\x12\t\n\x05\x42OOLS\x10\x66\"\x89\x02\n\x0bTensorProto\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x03\x12\x11\n\tdata_type\x18\x02 \x01(\x05\x12\x16\n\nfloat_data\x18\x04 \x03(\x02\x42\x02\x10\x01\x12\x16\n\nint32_data\x18\x05 \x03(\x05\x42\x02\x10\x01\x12\x13\n\x0bstring_data\x18\x06 \x03(\x0c\x12\x16\n\nint64_data\x18\x07 \x03(\x03\x42\x02\x10\x01\"|\n\x08\x44\x61taType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\t\n\x05UINT8\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06UINT16\x10\x04\x12\t\n\x05INT16\x10\x05\x12\t\n\x05INT32\x10\x06\x12\t\n\x05INT64\x10\x07\x12\n\n\x06STRING\x10\x08\x12\x08\n\x04\x42OOL\x10\t\"\x96\x01\n\x10TensorShapeProto\x12.\n\x03\x64im\x18\x01 \x03(\x0b\x32!.aihwx.TensorShapeProto.Dimension\x1aR\n\tDimension\x12\x13\n\tdim_value\x18\x01 \x01(\x03H\x00\x12\x13\n\tdim_param\x18\x02 \x01(\tH\x00\x12\x12\n\ndenotation\x18\x03 \x01(\tB\x07\n\x05value')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'onnx_common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TENSORPROTO.fields_by_name['float_data']._options = None
  _TENSORPROTO.fields_by_name['float_data']._serialized_options = b'\020\001'
  _TENSORPROTO.fields_by_name['int32_data']._options = None
  _TENSORPROTO.fields_by_name['int32_data']._serialized_options = b'\020\001'
  _TENSORPROTO.fields_by_name['int64_data']._options = None
  _TENSORPROTO.fields_by_name['int64_data']._serialized_options = b'\020\001'
  _ATTRIBUTEPROTO._serialized_start=29
  _ATTRIBUTEPROTO._serialized_end=430
  _ATTRIBUTEPROTO_ATTRIBUTETYPE._serialized_start=287
  _ATTRIBUTEPROTO_ATTRIBUTETYPE._serialized_end=430
  _TENSORPROTO._serialized_start=433
  _TENSORPROTO._serialized_end=698
  _TENSORPROTO_DATATYPE._serialized_start=574
  _TENSORPROTO_DATATYPE._serialized_end=698
  _TENSORSHAPEPROTO._serialized_start=701
  _TENSORSHAPEPROTO._serialized_end=851
  _TENSORSHAPEPROTO_DIMENSION._serialized_start=769
  _TENSORSHAPEPROTO_DIMENSION._serialized_end=851
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Configurations for resistive processing units."""

from .configs import (
    FloatingPointRPUConfig, InferenceRPUConfig, SingleRPUConfig,
    UnitCellRPUConfig, DigitalRankUpdateRPUConfig
)
from .enums import (
    BoundManagementType, NoiseManagementType,
    WeightNoiseType, PulseType, WeightModifierType, WeightClipType,
    WeightRemapType, VectorUnitCellUpdatePolicy, AnalogMVType
)
from .utils import (
    IOParameters, UpdateParameters,
    WeightModifierParameter, WeightClipParameter,
    WeightRemapParameter, SimpleDriftParameter, DriftParameter,
    MappingParameter, InputRangeParameter, PrePostProcessingParameter
)
from .devices import (
    FloatingPointDevice, IdealDevice, ConstantStepDevice,
    LinearStepDevice, SoftBoundsDevice, SoftBoundsPmaxDevice,
    SoftBoundsReferenceDevice, ExpStepDevice, PowStepDevice,
    PowStepReferenceDevice, PiecewiseStepDevice
)
from .compounds import (
    VectorUnitCell, ReferenceUnitCell,
    OneSidedUnitCell, DifferenceUnitCell, TransferCompound,
    BufferedTransferCompound, MixedPrecisionCompound
)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analog mapped layers."""

from typing import Optional, Tuple, List

from torch import Tensor, cat, split, no_grad
from torch.nn import Linear

from aihwkit.nn.functions import AnalogFunction
from aihwkit.nn.modules.base import AnalogModuleBase, RPUConfigAlias
from aihwkit.simulator.configs import SingleRPUConfig
from aihwkit.exceptions import ModuleError


class AnalogLinearMapped(AnalogModuleBase, Linear):
    """Linear layer that uses an analog tile.

    Linear layer that uses an analog tile during its forward, backward
    and update passes. In contrast to
    :class:`~aihwkit.bb.modules.linear.Linear` the maximal in and/or
    out dimension can be restricted, in which case the linear layer is
    split into multiple parts and computed on multiple tiles of given
    max sizes.

    In contrast to :class:`~aihwkit.bb.modules.linear.Linear`, the
    bias vector (if requested) is always handled in digital (floating
    point).

    Note:
        Mapping is controlled by the :class:`aihwkit.simulator.configs.utils.MappingParameter`.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_features: input vector size (number of columns).
        out_features: output vector size (number of rows).
        rpu_config: resistive processing unit configuration.
        bias: whether to use a bias row on the analog tile or not
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling
    """
    # pylint: disable=abstract-method, too-many-locals, too-many-instance-attributes

    __constants__ = ['in_features', 'out_features', 'realistic_read_write',
                     'digital_bias', 'analog_bias', 'use_bias']
    in_features: int
    out_features: int
    realistic_read_write: bool
    digital_bias: bool
    analog_bias: bool
    use_bias: bool
    in_sizes: List[int]
    out_sizes: List[int]

    def __init__(
            self,
            in_features: int,
            out_features: int,
            bias: bool = True,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None,
    ):

        # Call super() after tile creation, including ``reset_parameters``.
        Linear.__init__(self, in_features, out_features, bias=bias)

        # Create tiles
        if rpu_config is None:
            rpu_config = SingleRPUConfig()

        rpu_config = self._set_weight_scaling_omega(rpu_config, weight_scaling_omega)

        AnalogModuleBase.__init__(
            self,
            in_features,
            out_features,
            bias,
            realistic_read_write,
            rpu_config.mapping
        )
        if self.analog_bias:
            raise ModuleError("AnalogLinearMapped only supports digital bias.")

        # More than one tile may need to be created. If so, divide
        # weight matrix into equal pieces along input dimension with
        # as many tiles as needed
        max_input_size = rpu_config.mapping.max_input_size
        max_output_size = rpu_config.mapping.max_output_size

        self.in_sizes = self.get_split_sizes(in_features, max_input_size)
        self.out_sizes = self.get_split_sizes(out_features, max_output_size)

        self.analog_tile_array = []
        for i, in_tile_size in enumerate(self.in_sizes):
            in_tiles = []
            for j, out_tile_size in enumerate(self.out_sizes):
                tile = rpu_config.tile_class(out_tile_size,
                                             in_tile_size,
                                             rpu_config,
                                             bias=self.analog_bias)
                self.register_analog_tile(tile, name=f"{i}_{j}")
                in_tiles.append(tile)
            self.analog_tile_array.append(in_tiles)

        # Set weights from the reset_parameters
        self.set_weights(self.weight, self.bias)

        # Unregister weight/bias as a parameter but keep for sync
        self.unregister_parameter('weight')

    def get_split_sizes(self, size: int, split_max_size: int) -> List[int]:
        """ Computed the split sizes.

        Args:
            size: number of elements of the layer in one dimension
            split_max_size: max size of the split

        Returns:
            List of split sizes
        """
        if split_max_size <= 0:
            return [size]

        n_splits = (size + split_max_size - 1) // split_max_size
        base, extra = divmod(size, n_splits)
        return [base + (i < extra) for i in range(n_splits)]

    def set_weights(
            self,
            weight: Tensor,
            bias: Optional[Tensor] = None,
            force_exact: bool = False,
            apply_weight_scaling: bool = True,
            weight_scaling_omega: Optional[float] = None
    ) -> None:
        """Set the weight (and bias) with given Tensors.

        This uses an realistic read if the property ``realistic_read_write`` of
        the layer is set, unless it is overwritten by ``force_exact``. It
        scales the analog weights by the digital alpha scale if
        ``weight_scaling_omega`` is positive (see
        :meth:`~aihwkit.simulator.tiles.base.apply_weight_scaling`).

        Note:
            This is the recommended way for setting the weight/bias matrix of
            the analog tile, as it will correctly store the weights into the
            internal memory. Directly writing to ``self.weight`` and
            ``self.bias`` might yield wrong results as they are not always in
            sync with the analog tile Parameters, for performance reasons.

        Args:
            weight: weight matrix
            bias: bias vector
            force_exact: forces an exact write to the analog tiles
            apply_weight_scaling: Whether to rescale the given weight matrix
                and populate the digital output scaling factors as
                specified in the configuration
                :class:`~aihwkit.configs.utils.MappingParameter`. A
                new ``weight_scaling_omega`` can be given. Note that
                this will overwrite the existing digital out scaling
                factors.
            weight_scaling_omega: The weight scaling omega factor (see
                :class:`~aihwkit.configs.utils.MappingParameter`). If
                given explicitly here, it will overwrite the value in
                the mapping field.
        """
        shape = [self.out_features, self.in_features]
        weight = weight.clone().reshape(shape)

        realistic = self.realistic_read_write and not force_exact
        in_start = in_end = 0
        for in_size, in_tiles in zip(self.in_sizes, self.analog_tile_array):
            in_end += in_size
            out_start = out_end = 0
            for out_size, analog_tile in zip(self.out_sizes, in_tiles):
                out_end += out_size

                tile_weight = weight[out_start:out_end, in_start:in_end]

                analog_tile.set_weights(
                    tile_weight, None,
                    apply_weight_scaling,
                    weight_scaling_omega)

                if realistic:
                    analog_tile.program_weights()

                out_start = out_end
            in_start = in_end

        if self.digital_bias and bias is not None:
            with no_grad():
                self.bias.data[:] = bias[:]

        self._sync_weights_from_tile()

    def get_weights(self, force_exact: bool = False,
                    apply_weight_scaling: bool = True) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the weight (and bias) tensors.

        This uses an realistic read if the property ``realistic_read_write`` of
        the layer is set, unless it is overwritten by ``force_exact``. It
        scales the analog weights by the digital alpha scale if
        ``weight_scaling_omega`` is positive (see
        :meth:`~aihwkit.simulator.tiles.base.apply_weight_scaling`).

        Note:
            This is the recommended way for setting the weight/bias matrix from
            the analog tile, as it will correctly fetch the weights from the
            internal memory. Accessing ``self.weight`` and ``self.bias`` might
            yield wrong results as they are not always in sync with the
            analog tile library, for performance reasons.

        Args:
            force_exact: forces an exact read to the analog tiles

            apply_weight_scaling: Whether to return the weights with the
                (digital) output scaling factors applied. Note the
                "logical" weights of the layer which the DNN is
                effectively using are those with the output scales
                applied. If ``apply_weight_scaling`` is set to False, then
                only the weight values that is programmed onto the
                crossbar array are returned, without applying the
                digital scales. Default is True.

        Returns:
            tuple: weight matrix, bias vector
        """
        realistic = self.realistic_read_write and not force_exact

        weight_lst = []
        for in_tiles in self.analog_tile_array:
            in_tile_weight = []
            for analog_tile in in_tiles:
                if realistic:
                    tile_weight, _ = analog_tile.read_weights(apply_weight_scaling)
                else:
                    tile_weight, _ = analog_tile.get_weights(apply_weight_scaling)
                in_tile_weight.append(tile_weight)
            weight_lst.append(cat(in_tile_weight, 0))

        weight = cat(weight_lst, 1)

        if self.digital_bias:
            with no_grad():
                return weight, self.bias.data.clone().detach().cpu()
        return weight, None

    def reset_parameters(self) -> None:
        """Reset the parameters (weight and bias)."""
        super().reset_parameters()
        if self.analog_tile_count():
            self.set_weights(self.weight, self.bias)

    def forward(self, x_input: Tensor) -> Tensor:
        """Compute the forward pass."""
        # pylint: disable=arguments-differ,arguments-renamed

        if self.analog_tile_count() == 1:
            analog_tile = self.analog_tile_array[0][0]
            out = AnalogFunction.apply(
                analog_tile.get_analog_ctx(), x_input,
                analog_tile.shared_weights, not self.training)

            out = analog_tile.apply_out_scaling(out, (-1, ))

            if self.digital_bias:
                return out + self.bias
            return out

        # mapped version
        last_dim = x_input.ndim - 1
        splits = split(x_input, self.in_sizes, dim=last_dim)
        result = None  # type: Tensor
        for idx, (x, in_tiles) in enumerate(zip(splits, self.analog_tile_array)):
            out_result = []

            for analog_tile in in_tiles:
                output = AnalogFunction.apply(
                    analog_tile.get_analog_ctx(), x,
                    analog_tile.shared_weights, not self.training)

                output = analog_tile.apply_out_scaling(output, (-1, ))
                out_result.append(output)

            if idx == 0:
                result = cat(out_result, last_dim)
            else:
                result.add_(cat(out_result, last_dim))

        # add bias to final result
        if self.digital_bias:
            return result.add_(self.bias)
        return result

    def extra_repr(self) -> str:
        """Set the extra representation of the module.

        Returns:
            A string with the extra representation.
        """
        output = AnalogModuleBase.extra_repr(self)
        output += ', mapping={}'.format((len(self.in_sizes), len(self.out_sizes)))

        return output

    @classmethod
    def from_digital(
            cls,
            module: Linear,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogLinearMapped':
        """Return an AnalogLinearMapped layer from a torch Linear layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogLinearMapped layer based on the digital Linear ``module``.
        """
        analog_module = cls(module.in_features,
                            module.out_features,
                            module.bias is not None,
                            rpu_config,
                            realistic_read_write,
                            )

        analog_module.set_weights(module.weight, module.bias)
        return analog_module



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines

"""Utility parameters for resistive processing units."""

from dataclasses import dataclass, field
from typing import ClassVar, Type, Union, Any, List, TYPE_CHECKING

from aihwkit.simulator.configs.helpers import _PrintableMixin
from aihwkit.simulator.rpu_base import devices, tiles
from aihwkit.simulator.configs.enums import (
    BoundManagementType, NoiseManagementType, WeightNoiseType, PulseType,
    WeightModifierType, WeightClipType, WeightRemapType,
    AnalogMVType
)

if TYPE_CHECKING:
    from aihwkit.nn.modules.linear import AnalogLinear
    from aihwkit.nn.modules.linear_mapped import AnalogLinearMapped


@dataclass
class IOParameters(_PrintableMixin):
    """Parameter that define the analog-matvec (forward / backward) and
    peripheral digital input-output behavior.

    Here one can enable analog-digital conversion, dynamic input
    scaling, and define the properties of the analog-matvec
    computations, such as noise and non-idealities (e.g. IR-drop).
    """

    bindings_class: ClassVar[Type] = devices.AnalogTileInputOutputParameter

    is_perfect: bool = False
    """Short-cut to compute a perfect forward pass.

    If ``True``, it assumes an ideal forward pass (e.g. no bound, ADC etc...).
    Will disregard all other settings in this case.
    """

    mv_type: AnalogMVType = AnalogMVType.ONE_PASS
    """Selects the type of analog mat-vec computation. See
    :class:`AnalogMVType` for details. """

    inp_bound: float = 1.0
    """Input bound and ranges for the digital-to-analog converter (DAC)."""

    inp_noise: float = 0.0
    r"""Std deviation of Gaussian input noise (:math:`\sigma_\text{inp}`).

    i.e. noisiness of the analog input (at the stage after DAC and
    before the multiplication).
    """

    inp_res: float = 1 / (2**7 - 2)
    r"""Number of discretization steps for DAC (:math:`\le0` means infinite steps)
    or resolution (1/steps)."""

    inp_sto_round: bool = False
    """Whether to enable stochastic rounding of DAC."""

    inp_asymmetry: float = 0.0
    """Input asymmetry :math:`a_\text{input}`.

    Input of the negative input pass is scaled by :math:`(1 - a_\text{input})`.

    Note:
        This setting has only effect in case of and
        :class:`AnalogMVType` that uses separate passes for positive
        and negative inputs.
    """

    out_bound: float = 12.0
    """Output bound and ranges for analog-to-digital converter (ADC)."""

    out_noise: float = 0.06
    r"""Std deviation of Gaussian output noise (:math:`\sigma_\text{out}`).

    i.e. noisiness of device summation at the output.
    """

    out_res: float = 1 / (2**9 - 2)
    """Number of discretization steps for ADC or resolution.

    Number of discretization steps for ADC (:math:`<=0` means infinite steps)
    or resolution (1/steps).
    """

    out_sto_round: bool = False
    """Whether to enable stochastic rounding of ADC."""

    out_scale: float = 1.0
    """Additional fixed scalar factor."""

    out_asymmetry: float = 0.0
    """Output asymmetry :math:`a_\text{output}`.

    Output of the negative input pass is scaled by :math:`(1 - a_\text{output})`.

    Note:
        This setting has only effect in case of and
        :class:`AnalogMVType` that uses separate passes for positive
        and negative inputs.
    """

    bound_management: BoundManagementType = BoundManagementType.ITERATIVE

    """Type of bound management, see :class:`BoundManagementType`.

    Caution:
        Bound management is **only** available for the forward pass. It
        will be ignored when used for the backward pass.
    """

    noise_management: NoiseManagementType = NoiseManagementType.ABS_MAX
    """Type of noise management, see :class:`NoiseManagementType`."""

    w_noise: float = 0.0
    r"""Scale of output referred weight noise (:math:`\sigma_w`) for a given
    ``w_noise_type``."""

    w_noise_type: WeightNoiseType = WeightNoiseType.NONE
    """Type as specified in :class:`OutputWeightNoiseType`.

    Note:
        This noise us applied each time anew as it is referred to
        the output. It will not change the conductance values of
        the weight matrix. For the latter one can apply
        :meth:`diffuse_weights`.
    """

    ir_drop: float = 0.0
    """Scale of IR drop along the inputs (rows of the weight matrix).

    The IR-drop is calculated assuming that the first input is
    farthest away from the output channel. The expected drop is
    approximating the steady-state voltage distributions and depends
    on the input current.
    """

    ir_drop_g_ratio: float = 1.0 / 0.35 / 5e-6
    """Physical ratio of wire conductance from one cell to the next to
    physical max conductance of a device.

    Default is compute with 5mS maximal conductance set state and 0.35
    Ohm wire resistance.
    """

    out_nonlinearity: float = 0.0
    """S-shaped non-linearity applied to the analog output.

    Output non-linearity applies an S-shaped non-linearity to the
    analog output (before the ADC), i.e. :math:`\frac{y_i}{1 +
    n_i*|y_i|}` where :math:`n_i` is drawn at the instantiation time
    by::
        out_nonlinearity / out_bound * (1 + out_nonlinearity_std * rand)
    """

    out_nonlinearity_std: float = 0.0
    """ Output-to-output non linearity variation. """

    slope_calibration: float = 0.0
    """Models a calibration process of the output non-linearity (and
    r-series).

    This is the relative value in the output range where the slope of
    the non-linearity should have slope 1. E.g. 0.5 would be at half-out
    range.
    """

    v_offset_std: float = 0.0
    """Voltage offset variation.

    The output is multiplied by a systematic factor set for each
    output line at time of instantiation, e.g. :math:`(1 - v_i)` for
    the coding device and :math:`(1 + v_i)` for the reference device
    (assuming differential reads).

    """

    v_offset_w_min: float = -1.0
    """ Voltage offset for an implicit reference unit. """

    r_series: float = 0.0
    """Series resistance in fraction of the total output current."""

    w_read_asymmetry_dtod: float = 0.0
    """Device polarity read dependence.

    The negative inputs perceive a slightly different weight (e.g. pcm
    polarity dependence). Each device has a different factor, and the
    spread of this device-to-device variability can be set with
    ``w_read_asymmetry_dtod``. A weight (given negative input) will be
    then scaled by :math:`1 - f_{ij}` where :math:`f_{ij}` is drawn
    from a Gaussian distribution (with zero mean and standard
    deviation ``w_read_asymmetry_dtod``).
    """

    max_bm_factor: int = 1000

    """Maximal bound management factor.

    If this factor is reached then the iterative process is stopped.
    """

    max_bm_res: float = 0.25
    """Limit the maximal number of iterations of the bound management.

    Another way to limit the maximal number of iterations of the bound
    management. The max effective resolution number of the inputs, e.g. use
    :math:`1/4` for 2 bits.
    """
    bm_test_negative_bound: bool = True

    nm_thres: float = 0.0
    r"""Constant noise management value for ``type`` ``Constant``.

    In other cases, this is a upper threshold :math:`\theta` above which the
    noise management factor is saturated. E.g. for `AbsMax`:

    .. math::
        :nowrap:

        \begin{equation*} \alpha=\begin{cases}\max_i|x_i|, &
        \text{if} \max_i|x_i|<\theta \\ \theta, &
        \text{otherwise}\end{cases} \end{equation*}

    Caution:
        If ``nm_thres`` is set (and type is not ``Constant``), the noise
        management will clip some large input values, in favor of having a
        better SNR for smaller input values.
    """


@dataclass
class UpdateParameters(_PrintableMixin):
    """Parameter that modify the update behaviour of a pulsed device."""

    bindings_class: ClassVar[Type] = devices.AnalogTileUpdateParameter

    desired_bl: int = 31
    """Desired length of the pulse trains.

    For update BL management, it is the maximal pulse train length.
    """

    fixed_bl: bool = True
    """Whether to fix the length of the pulse trains.

    See also ``update_bl_management``.

    In case of ``True`` (where ``dw_min`` is the mean minimal weight change
    step size) it is::

        BL = desired_BL
        A = B =  sqrt(learning_rate / (dw_min * BL))

    In case of ``False``::

        if dw_min * desired_BL < learning_rate:
            A = B = 1
            BL = ceil(learning_rate / dw_min
        else:
            # same as for fixed_BL=True
    """

    pulse_type: PulseType = PulseType.STOCHASTIC_COMPRESSED
    """Switching between different pulse types.

    See also :class:`PulseTypeMap` for details.

    Important:
        Pulsing can also be turned off in which case the update is done as if
        in floating point and all other update related parameter are ignored.
    """

    res: float = 0
    """Resolution of the update probability for the stochastic bit line
    generation.

    Resolution ie. bin width in ``0..1``) of the update probability for the
    stochastic bit line generation. Use -1 for turning discretization off. Can
    be given as number of steps as well.
    """

    x_res_implicit: float = 0
    """Resolution of each quantization step for the inputs ``x``.

    Resolution (ie. bin width) of each quantization step for the inputs ``x``
    in case of ``DeterministicImplicit`` pulse trains. See
    :class:`PulseTypeMap` for details.
    """

    d_res_implicit: float = 0
    """Resolution of each quantization step for the error ``d``.

    Resolution (ie. bin width) of each quantization step for the error ``d``
    in case of `DeterministicImplicit` pulse trains. See
    :class:`PulseTypeMap` for details.
    """

    sto_round: bool = False
    """Whether to enable stochastic rounding."""

    update_bl_management: bool = True
    """Whether to enable dynamical adjustment of ``A``,``B``,and ``BL``::

        BL = ceil(learning_rate * abs(x_j) * abs(d_i) / weight_granularity);
        BL  = min(BL,desired_BL);
        A = B = sqrt(learning_rate / (weight_granularity * BL));

    The ``weight_granularity`` is usually equal to ``dw_min``.
    """

    update_management: bool = True
    r"""Whether to apply additional scaling.

    After the above setting an additional scaling (always on when using
    `update_bl_management``) is applied to account for the different input
    strengths.
    If

    .. math:: \gamma \equiv \max_i |x_i| / (\alpha \max_j |d_j|)

    is the ratio between the two maximal inputs, then ``A`` is additionally
    scaled by :math:`\gamma` and ``B`` is scaled by :math:`1/\gamma`.

    The gradient scale :math:`\alpha` can be set with ``um_grad_scale``
    """

    um_grad_scale: float = 1.0
    r"""Scales the gradient for the update management.

    The factor :math:`\alpha` for the ``update_management``. If
    smaller than 1 it means that the gradient will be earlier clipped
    when learning rate is too large (ie. exceeding the maximal
    pulse number times the weight granularity). If 1, both d and x inputs
    are clipped for the same learning rate.
    """


@dataclass
class WeightModifierParameter(_PrintableMixin):
    """Parameter that modify the forward/backward weights during hardware-aware training."""

    bindings_class: ClassVar[Type] = tiles.WeightModifierParameter

    std_dev: float = 0.0
    """Standard deviation of the added noise to the weight matrix.

    This parameter affects the modifier types ``AddNormal``, ``MultNormal`` and
    ``DiscretizeAddNormal``.

    Note:
        If the parameter ``rel_to_actual_wmax`` is set then the ``std_dev`` is
        computed in relative terms to the abs max of the given weight matrix,
        otherwise it in relative terms to the assumed max, which is set by
        ``assumed_wmax``.
    """

    res: float = 0.0
    r"""Resolution of the discretization.

    The invert of ``res`` gives the number of equal sized steps in
    :math:`-a_\text{max}\ldots,a_\text{max}` where the
    :math:`a_\text{max}` is either given by the abs max (if
    ``rel_to_actual_wmax`` is set) or ``assumed_wmax`` otherwise.

    ``res`` is only used in the modifier types ``DoReFa``, ``Discretize``, and
    ``DiscretizeAddNormal``.
    """

    sto_round: bool = False
    """Whether the discretization is done with stochastic rounding enabled.

    ``sto_round`` is only used in the modifier types ``DoReFa``,
    ``Discretize``, and ``DiscretizeAddNormal``.
    """

    dorefa_clip: float = 0.6
    """Parameter for DoReFa."""

    pdrop: float = 0.0
    """Drop connect probability.

    Drop connect sets weights to zero with the given probability. This
    implements drop connect.

    Important:
        Drop connect can be used with any other modifier type in combination.
    """

    enable_during_test: bool = False
    """Whether to use the last modified weight matrix during testing.

    Caution:
        This will **not** remove drop connect or any other noise
        during evaluation, and thus should only used with care.
    """

    rel_to_actual_wmax: bool = True
    """Whether to calculate the abs max of the weight and apply noise relative
    to this number.

    If set to False, ``assumed_wmax`` is taken as relative units.
    """

    assumed_wmax: float = 1.0
    """Assumed weight value that is mapped to the maximal conductance.

    This is typically 1.0. This parameter will be ignored if
    ``rel_to_actual_wmax`` is set.
    """

    copy_last_column: bool = False
    """Whether to not apply noise to the last column (which usually contains
    the bias values)."""

    coeffs: List[float] = field(
        default_factory=lambda: [0.0105392, 0.0768, -0.046925],
        metadata={'hide_if': [0.0105392, 0.0768, -0.046925]}
    )

    """Coefficients for the ``POLY`` weight modifier type.

    See :class:`WeightModifierType` for details.
    """

    type: WeightModifierType = WeightModifierType.COPY
    """Type of the weight modification."""


@dataclass
class WeightClipParameter(_PrintableMixin):
    """Parameter that clip the weights during hardware-aware training.

    Important:
        A clipping ``type`` has to be set before any of the parameter
        changes take any effect.

    """

    bindings_class: ClassVar[Type] = tiles.WeightClipParameter

    fixed_value: float = -1.0
    """Clipping value in case of ``FixedValue`` type.

    Caution:

        If ``fixed_value > 0`` it will be also applied during other
        clipping types.

    """

    sigma: float = 2.5
    """Sigma value for clipping for the ``LayerGaussian`` type."""

    type: WeightClipType = WeightClipType.NONE
    """Type of clipping."""


@dataclass
class WeightRemapParameter(_PrintableMixin):
    """Parameter that remap the weights during hardware-aware training.

    Important:
        A remap ``type`` has to be set before any of the parameter
        changes take any effect.
    """
    bindings_class: ClassVar[Type] = tiles.WeightRemapParameter

    remapped_wmax: float = 1.0
    """Assumed max of weight, ie the value of the weight the maximal
    conductance is mapped to. Typically 1.0.
    """

    max_scale_range: float = 0.0
    """Maximal range of scale values. Use zero to turn any restrictions
    off (default)."""

    max_scale_ref: float = 0.0
    """Reference scale that use used as minimal scale for determining the
    scale range."""

    type: WeightRemapType = WeightRemapType.NONE
    """Type of clipping."""


@dataclass
class SimpleDriftParameter(_PrintableMixin):
    r"""Parameter for a simple power law drift.

    The drift as a simple power law drift without device-to-device
    variation or conductance dependence.

    It computes:
    .. math::

        w_{ij}*\left(\frac{t + \Delta t}{t_0}\right)^(-\nu)
    """

    bindings_class: ClassVar[Type] = devices.DriftParameter

    nu: float = 0.0
    r"""Average drift :math:`\nu` value.

    Need to non-zero to actually use the drift.
    """

    t_0: float = 1.0
    """Time between write and first read.

    Usually assumed in milliseconds, however, it really determines the time
    units of ``time_since_last_call`` when calling the drift.
    """

    reset_tol: float = 1e-7
    """Reset tolerance.

    This should a number smaller than the expected weight change as it is used
    to detect any changes in the weight from the last drift call. Every change
    to the weight above this tolerance will reset the drift time.

    Caution:
        Any write noise or diffusion on the weight might thus
        interfere with the drift.
   """


@dataclass
class DriftParameter(SimpleDriftParameter):
    r"""Parameter for a power law drift.

    The drift is based on the model described by `Oh et al (2019)`_.

    It computes:
    .. math::

        w_{ij}*\left(\frac{t + \Delta t}{t_0}\right)^(-\nu^\text{actual}_{ij})

    where the drift coefficient is drawn once at the beginning and
    might depend on device. It also can depend on the actual weight
    value.

    The actual drift coefficient is computed as:
    .. math::

        \nu_{ij}^\text{actual} =  \nu_{ij} - \nu_k \log \frac{(w_{ij} - w_\text{off}) / r_\text{wg}
        + g_\text{off}}{G_0}  + \nu\sigma_\nu\xi

    here :math:`w_{ij}` is the actual weight and `\nu_{ij}` fixed for
    each device given by the mean :math:`\nu` and the device-to-device
    variation: :math:`\nu_{ij} = \nu + \nu_dtod\nu\xi` and are only
    drawn once at the beginning (tile instantiation).  `\xi` is
    Gaussian noise.

    Note:
        If the weight has changed from the last drift call (determined
        by the ``reset_tol`` parameter), for instance due to update,
        decay or noise, then the drift time :math:`t` will be reset and start
        from new, however, the drift coefficients :math:`\nu_{ij}` are
        *not* changed. On the other hand, if the weights has not
        changed since last call, :math:`t` will accumulate the time.

    Caution:
        Note that the drift coefficient does *not* depend on the initially
        programmed weight value at :math:`t=0` in the current
        implementation (ie G0 is a constant for all devices), but
        instead on the actual weight. In some materials (e.g. phase
        changed materials), that might be not accurate.

    .. _`Oh et al (2019)`: https://ieeexplore.ieee.org/document/8753712
    """

    bindings_class: ClassVar[Type] = devices.DriftParameter

    nu_dtod: float = 0.0
    r"""Device-to-device variation of the :math:`\nu` values."""

    nu_std: float = 0.0
    r"""Cycle-to-cycle variation of :math:`\nu`.

    A more realistic way to add noise of the drift might be using
    ``w_noise_std``.
    """

    wg_ratio: float = 1.0
    """``(w_max-w_min)/(g_max-g_min)`` to convert to physical units."""

    g_offset: float = 0.0
    """``g_min`` to convert to physical units."""

    w_offset: float = 0.0
    """``w(g_min)``, i.e. to what value ``g_min`` is mapped to in w-space."""

    nu_k: float = 0.0
    r"""Variation of math:`nu` with :math:`W`.

    That is :math:`\nu(R) = nu_0 - k \log(G/G_0)`.  See Oh et al. for
    details.
    """

    log_g0: float = 0.0
    """Log g0."""

    w_noise_std: float = 0.0
    """Additional weight noise (Gaussian diffusion) added to the weights
    after the drift is applied."""


@dataclass
class MappingParameter(_PrintableMixin):
    """Parameter related to hardware design and the mapping of logical
    weight matrices to physical tiles.

    Caution:

        Some of these parameters have only an effect for modules that
        support tile mappings.
    """

    digital_bias: bool = True
    """Whether the bias term is handled by the analog tile or kept in
    digital.

    Note:
        Default is having a *digital* bias so that bias values are
        *not* stored onto the analog crossbar. This needs to be
        supported by the chip design. Set to False if the analog bias
        is instead situated on the the crossbar itself (as an extra
        column)

    Note:
        ``digital_bias`` is supported by *all* analog modules.
    """

    weight_scaling_omega: float = 0.0
    """omega_scale is a user defined parameter used to scale the weights
    while remapping these to cover the full range of values allowed.
    """

    weight_scaling_columnwise: bool = False
    """Whether the weight matrix will be remapped column-wise over
    the maximum device allowed value."""

    learn_out_scaling: bool = False
    """Define (additional) out scales that are learnable parameter
    used to scale the output."""

    out_scaling_columnwise: bool = False
    """Whether the learnable out scaling parameter enabled by
    ``learn_out_scaling`` is a scalar (``False``) or learned for
    each output (``True``).
    """

    max_input_size: int = 512
    """Maximal input size (number of columns) of the weight matrix
    that is handled on a single analog tile.

    If the logical weight matrix size exceeds this size it will be
    split and mapped onto multiple analog tiles.

    Caution:
        Only relevant for ``Mapped`` modules such as
        :class:`aihwkit.nn.modules.linear_mapped.AnalogLinearMapped`.
    """

    max_output_size: int = 512
    """Maximal output size (number of rows) of the weight matrix
    that is handled on a single analog tile.

    If the logical weight matrix size exceeds this size it will be
    split and mapped onto multiple analog tiles.

    Caution:
        Only relevant for ``Mapped`` modules such as
        :class:`aihwkit.nn.modules.linear_mapped.AnalogLinearMapped`.
    """


@dataclass
class InputRangeParameter(_PrintableMixin):
    """ Parameter related to input range learning """

    enable: bool = False
    """Whether to enable to learn the input range. Note that if enable is
    ``False`` then no clip is applied.

    Note:

        The input bound (``forward.inp_bound``) is assumed to be 1 if
        enabled as the input range already scales the input into to the
        range :math:`(-1, 1)` by dividing the input to the type by
        itself and multiplying the output accordingly.

        Typically, noise and bound management should be set to `NONE`
        for the input range learning as it replaces the dynamic
        managements with a static but learned input bound. However, in
        some exceptional experimental cases one might want to enable
        the management techniques on top of the input range learning,
        so that no error is raised if they are not set to `NONE`.
    """

    init_value: float = 3.0
    """Initial setting of the input range in case of input range learning."""

    init_from_data: int = 100
    """Number of batches to use for initialization from data. Set 0 to turn off."""

    init_std_alpha: float = 3.0
    """Standard deviation multiplier for initialization from data."""

    decay: float = 0.001
    """Decay rate for input range learning."""

    input_min_percentage: float = 0.95
    """Decay is only applied if percentage of non-clipped values is above this value.

    Note:

        The added gradient is (in case of non-clipped input
        percentage ``percentage > input_min_percentage``)::

            grad += decay * input_range
    """

    manage_output_clipping: bool = True
    """Whether to increase the input range when output clipping occurs.

    Caution:

        The output bound is taken from the ``forward.out_bound``
        value, which has to exist. Noise and bound management have to
        be set to NONE if this feature is enabled otherwise a
        ``ConfigError`` is raised.

    """

    output_min_percentage: float = 0.95
    """Increase of the input range is only applied if percentage of
    non-clipped output values is below this value.

    Note:

        The gradient subtracted from the input range is (in case of
        ``output_percentage < output_min_percentage``)::

            grad -= (1.0 - output_percentage) * input_range
    """

    gradient_scale: float = 1.0
    """Scale of the gradient magnitude (learning rate) for the input range learning."""

    gradient_relative: bool = True
    """Whether to make the gradient of the input range learning relative to
    the current range value.
    """

    def supports_manage_output_clipping(self, rpu_config: Any) -> bool:
        """ Checks whether rpu_config supported ``manage_output_clipping``.

        Args:
            rpu_config: RPUConfig to check

        Returns:
            True if supported otherwise False
        """

        if not hasattr(rpu_config, 'forward') or rpu_config.forward.is_perfect:
            return False
        if not isinstance(rpu_config.forward, IOParameters):
            return False
        if rpu_config.forward.noise_management != NoiseManagementType.NONE:
            return False
        if rpu_config.forward.bound_management != BoundManagementType.NONE:
            return False
        return True


@dataclass
class PrePostProcessingParameter(_PrintableMixin):
    """Parameter related to digital input and output processing, such as input clip
    learning.
     """
    input_range: InputRangeParameter = field(default_factory=InputRangeParameter)


@dataclass
class MapableRPU(_PrintableMixin):
    """Defines the mapping parameters and utility factories"""

    mapping: MappingParameter = field(default_factory=MappingParameter)
    """Parameter related to mapping weights to tiles for supporting modules."""

    def get_linear(self) -> Union[Type['AnalogLinear'], Type['AnalogLinearMapped']]:
        """Returns a AnalogLinear module as specified """
        # pylint: disable=import-outside-toplevel
        # need to import here to avoid circular imports
        from aihwkit.nn.modules.linear import AnalogLinear
        from aihwkit.nn.modules.linear_mapped import AnalogLinearMapped

        if self.mapping.max_input_size > 0 or self.mapping.max_output_size > 0:
            return AnalogLinearMapped
        return AnalogLinear


@dataclass
class PrePostProcessingRPU(_PrintableMixin):
    """Defines the pre-post parameters and utility factories"""

    pre_post: PrePostProcessingParameter = field(default_factory=PrePostProcessingParameter)
    """Parameter related digital pre and post processing."""



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Legacy class for import the noise models for inference"""

import warnings
from aihwkit.inference import *  # pylint: disable=unused-wildcard-import, wildcard-import

warnings.warn(
    FutureWarning("\n\nThe `aihwkit.simulator.noise_models` module has been superseded "
                  "by the `aihwkit.inference` module. "
                  "Please replace `from aihwkit.simulator.noise_models import ...` "
                  "with `from aihwkit.inference import ...` in your import statement.\n"))



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Neural network modules."""

# Convenience imports for easier access to the classes.

from aihwkit.nn.modules.container import AnalogSequential
from aihwkit.nn.modules.conv import AnalogConv1d, AnalogConv2d, AnalogConv3d
from aihwkit.nn.modules.linear import AnalogLinear
from aihwkit.nn.modules.rnn.rnn import AnalogRNN
from aihwkit.nn.modules.rnn.cells import AnalogGRUCell, AnalogLSTMCell, AnalogVanillaRNNCell, \
    AnalogLSTMCellCombinedWeight
from aihwkit.nn.modules.linear_mapped import AnalogLinearMapped
from aihwkit.nn.modules.conv_mapped import (
    AnalogConv1dMapped, AnalogConv2dMapped, AnalogConv3dMapped
)




# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level inference tools."""

# Convenience imports for easier access to the classes.
from aihwkit.inference.converter.base import BaseConductanceConverter
from aihwkit.inference.converter.conductance import SinglePairConductanceConverter
from aihwkit.inference.noise.base import BaseNoiseModel
from aihwkit.inference.noise.pcm import PCMLikeNoiseModel
from aihwkit.inference.noise.custom import StateIndependentNoiseModel
from aihwkit.inference.compensation.base import BaseDriftCompensation
from aihwkit.inference.compensation.drift import GlobalDriftCompensation
from aihwkit.inference.utils import drift_analog_weights, program_analog_weights



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Configuration for NVM devices for Analog (Resistive Device) tiles."""

# pylint: disable=too-many-instance-attributes, too-many-lines

from dataclasses import dataclass, field
from typing import ClassVar, List, Type
from numpy import exp

from aihwkit.simulator.configs.helpers import (
    _PrintableMixin, parameters_to_bindings
)
from aihwkit.simulator.configs.utils import (
    DriftParameter, SimpleDriftParameter
)
from aihwkit.simulator.rpu_base import devices

# legacy
from aihwkit.simulator.configs.compounds import (  # pylint: disable=unused-import
    VectorUnitCell, ReferenceUnitCell,
    OneSidedUnitCell, DifferenceUnitCell, TransferCompound,
    BufferedTransferCompound, MixedPrecisionCompound
)


@dataclass
class FloatingPointDevice(_PrintableMixin):
    """Floating point reference.

    Implements ideal devices forward/backward/update behavior.
    """

    bindings_class: ClassVar[Type] = devices.FloatingPointTileParameter

    diffusion: float = 0.0
    """Standard deviation of diffusion process."""

    lifetime: float = 0.0
    r"""One over `decay_rate`, ie :math:`1/r_\text{decay}`."""

    drift: SimpleDriftParameter = field(default_factory=SimpleDriftParameter)
    """Parameter governing a power-law drift."""

    def as_bindings(self) -> devices.FloatingPointTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return parameters_to_bindings(self)

    def requires_diffusion(self) -> bool:
        """Return whether device has diffusion enabled."""
        return self.diffusion > 0.0

    def requires_decay(self) -> bool:
        """Return whether device has decay enabled."""
        return self.lifetime > 0.0


@dataclass
class PulsedDevice(_PrintableMixin):
    r"""Pulsed update resistive devices.

    Device are used as part of an
    :class:`~aihwkit.simulator.tiles.AnalogTile` to implement the
    `update once` characteristics, i.e. the material response properties
    when a single update pulse is given (a coincidence between row and
    column pulse train happened).

    Common properties of all pulsed devices include:

    **Reset**:

    Resets the weight in cross points to (around) zero with
    cycle-to-cycle and systematic spread around a mean.

    Important:
        Reset with given parameters is only activated when
        :meth:`~aihwkit.simulator.tiles.base.Base.reset_weights` is
        called explicitly by the user.

    **Decay**:

    .. math:: w_{ij} \leftarrow (w_{ij} - b_{ij})\,(1-\alpha_\text{decay}\delta_{ij}) + b_{ij}

    Weight decay is only activated by inserting a specific call to
    :meth:`~aihwkit.simulator.tiles.base.Base.decay_weights`, which is
    done automatically for a tile each mini-batch is decay is
    present. Note that the device ``decay_lifetime`` parameters (1
    over decay rates :math:`\delta_{ij}`) are analog tile specific and
    are thus set and fixed during RPU
    initialization. :math:`\alpha_\text{decay}` is a scaling factor
    that can be given during run-time.

    The bias :math:`b_{ij}` is given by the reset bias and which is
    determined by the parameter ``reset`` (mean value) and
    ``reset_dtod``  (device-to-device variability). Thus

    .. math:: b_{ij} = \mu_\text{reset} \left(1 + \sigma_\text{reset-dtod}\xi\right)

    Note that the reset bias is also applied in case the device is
    reset (see above).


    **Diffusion**:

    Similar to the decay, diffusion is only activated by inserting a
    specific call to
    :meth:`~aihwkit.simulator.tiles.base.Base.diffuse_weights`, which is
    done automatically for a tile each mini-batch is diffusion is
    present. The parameters of the diffusion process are set during
    RPU initialization and are fixed for the remainder.

    .. math:: w_{ij} \leftarrow w_{ij} + \rho_{ij} \, \xi;

    where :math:`xi` is a standard Gaussian variable and :math:`\rho_{ij}` the
    diffusion rate for a cross-point `ij`.

    Note:
        If diffusion happens to move the weight beyond the hard bounds of the
        weight it is ensured to be clipped appropriately.

    **Drift**:

    Optional power-law drift setting, as described in
    :class:`~aihwkit.similar.configs.utils.DriftParameter`.

    Important:
        Similar to reset, drift is *not* applied automatically each
        mini-batch but requires an explicit call to
        :meth:`~aihwkit.simulator.tiles.base.Base.drift_weights` each
        time the drift should be applied.

    """

    bindings_class: ClassVar[Type] = devices.PulsedResistiveDeviceParameter

    construction_seed: int = 0
    """If not equal 0, will set a unique seed for hidden parameters during
    construction."""

    corrupt_devices_prob: float = 0.0
    """Probability for devices to be corrupt (weights fixed to random value
    with hard bounds, that is min and max bounds are set to equal)."""

    corrupt_devices_range: float = 0.1
    """Range around zero for establishing corrupt devices."""

    diffusion: float = 0.0
    """Standard deviation of diffusion process."""

    diffusion_dtod: float = 0.0
    """Device-to device variation of diffusion rate in relative units."""

    drift: DriftParameter = field(default_factory=DriftParameter,
                                  metadata={'hide_if': DriftParameter()})
    """Parameter governing a power-law drift."""

    dw_min: float = 0.001
    """Mean of the minimal update step sizes across devices and directions."""

    dw_min_dtod: float = 0.3
    """Device-to-device std deviation of ``dw_min`` (in relative units to
    ``dw_min``)."""

    dw_min_dtod_log_normal: bool = False
    """Device-to-device std deviation ``dw_min_dtod`` given using a
    log-normal instead of normal distribution."""

    dw_min_std: float = 0.3
    r"""Cycle-to-cycle variation size of the update step (related to
    :math:`\sigma_\text{c-to-c}` above) in relative units to ``dw_min``.

    Note:
        Many spread (device-to-device variation) parameters are given in
        relative units. For instance e.g. a setting of ``dw_min_std`` of 0.1
        would mean 10% spread around the mean and thus a resulting standard
        deviation (:math:`\sigma_\text{c-to-c}`) of ``dw_min`` * ``dw_min_std``.
    """

    enforce_consistency: bool = True
    """Whether to enforce weight bounds consistency during initialization.

    Whether to enforce that max weight bounds cannot be smaller than min
    weight bounds, and up direction step size is positive and down negative.
    Switches the opposite values if encountered during init.
    """

    lifetime: float = 0.0
    r"""One over `decay_rate`, ie :math:`1/r_\text{decay}`."""

    lifetime_dtod: float = 0.0
    """Device-to-device variation in the decay rate (in relative units)."""

    perfect_bias: bool = False
    """No up-down differences and device-to-device variability in the bounds
    for the devices in the bias row."""

    reset: float = 0.0
    """The reset values and spread per cross-point ``ij`` when using reset
    functionality of the device."""

    reset_dtod: float = 0.0
    """See ``reset``."""

    reset_std: float = 0.01
    """See ``reset``."""

    up_down: float = 0.0
    r"""Up and down direction step sizes can be systematically different and
    also vary across devices.

    :math:`\Delta w_{ij}^d` is set during RPU initialization (for each
    cross-point :math:`ij`):

    .. math::

        \Delta w_{ij}^d = d\; \Delta w_\text{min}\, \left(
        1 + d \beta_{ij} + \sigma_\text{d-to-d}\xi\right)

    where :math:`\xi` is again a standard Gaussian. :math:`\beta_{ij}` is the
    directional up `versus` down bias.  At initialization ``up_down_dtod`` and
    ``up_down`` defines this bias term:

    .. math::

        \beta_{ij} = \beta_\text{up-down} + \xi
        \sigma_\text{up-down-dtod}

    where :math:`\xi` is again a standard Gaussian number and
    :math:`\beta_\text{up-down}` corresponds to ``up_down``. Note that
    ``up_down_dtod`` is again given in relative units to ``dw_min``.
    """

    up_down_dtod: float = 0.01
    """See ``up_down``."""

    w_max: float = 0.6
    """See ``w_min``."""

    w_max_dtod: float = 0.3
    """See ``w_min_dtod``."""

    w_min: float = -0.6
    """Mean of hard bounds across device cross-point `ij`.

    The parameters ``w_min`` and ``w_max`` are used to set the min/max bounds
    independently.

    Note:
        For this abstract device, we assume that weights can have
        positive and negative values and are symmetrically around
        zero. In physical circuit terms, this might be implemented
        as a difference of two resistive elements.
    """

    w_min_dtod: float = 0.3
    """Device-to-device variation of the hard bounds.

    Device-to-device variation of the hard bounds, of min and max value,
    respectively. All are given in relative units to ``w_min``, or ``w_max``,
    respectively.
    """

    def as_bindings(self) -> devices.PulsedResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return parameters_to_bindings(self)

    def requires_diffusion(self) -> bool:
        """Return whether device has diffusion enabled."""
        return self.diffusion > 0.0

    def requires_decay(self) -> bool:
        """Return whether device has decay enabled."""
        return self.lifetime > 0.0


###############################################################################
# Specific devices based on ``pulsed``.
###############################################################################

@dataclass
class IdealDevice(_PrintableMixin):
    """Ideal update behavior (using floating point), but forward/backward
    might be non-ideal.

    Ideal update behavior (using floating point), however,
    forward/backward might still have a non-ideal ADC or noise added.
    """

    bindings_class: ClassVar[Type] = devices.IdealResistiveDeviceParameter

    construction_seed: int = 0
    """If not ``0``, set a unique seed for hidden parameters during
    construction."""

    diffusion: float = 0.0
    """Standard deviation of diffusion process."""

    lifetime: float = 0.0
    r"""One over `decay_rate`, ie :math:`1/r_\text{decay}`."""

    def as_bindings(self) -> devices.IdealResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return parameters_to_bindings(self)

    def requires_diffusion(self) -> bool:
        """Return whether device has diffusion enabled."""
        return self.diffusion > 0.0

    def requires_decay(self) -> bool:
        """Return whether device has decay enabled."""
        return self.lifetime > 0.0


@dataclass
class ConstantStepDevice(PulsedDevice):
    r"""Pulsed update behavioral model: constant step.

    Pulsed update behavioral model, where the update step of
    material is constant throughout the resistive range (up to hard
    bounds).

    In more detail, the update behavior implemented for ``ConstantStep``
    is:

    .. math::

        w_{ij}  &\leftarrow&  w_{ij} - \Delta w_{ij}^d(1 + \sigma_\text{c-to-c}\,\xi)

        w_{ij}  &\leftarrow& \text{clip}(w_{ij},b^\text{min}_{ij},b^\text{max}_{ij})

    where :math:`d` is the direction of the update (product of signs
    of input and error). :math:`\Delta w_{ij}^d` is the update step
    size of the cross-point `ij` in direction :math:`d` (up or down).
    Note that each cross-point has separate update sizes so that
    device-to-device fluctuations and biases in the directions can be
    given.

    Moreover, the clipping bounds of each cross-point `ij`
    (i.e. :math:`b_{ij}^\text{max/min}`) are also different in
    general. The mean and the amount of systematic spread from
    device-to-device can be given as parameters, see below.

    For parameters regarding the devices settings, see e.g.
    :class:`~PulsedDevice`.
    """

    bindings_class: ClassVar[Type] = devices.ConstantStepResistiveDeviceParameter


@dataclass
class LinearStepDevice(PulsedDevice):
    r"""Pulsed update behavioral model: linear step.

    Pulsed update behavioral model, where the update step response
    size of the material is linearly dependent with resistance (up to
    hard bounds).

    This model is based on :class:`~PulsedDevice` and thus
    shares all parameters and functionality. In addition, it only
    implements a more general `update once` function, where the update
    step size can depend linearly on the weight itself.

    For each coincidence the weights is updated once. Here, the
    positive (negative) update step size decreases linearly in the
    following manner (compare to the `update once` for
    :class:`~ConstantStepDevice`):

    .. math::
        :nowrap:

        \begin{eqnarray*}
        w_{ij}  &\leftarrow&  w_{ij} - \Delta w_{ij}^d(\gamma_{ij}^d\;w_{ij}
        + 1 + \sigma_\text{c-to-c}\,\xi)\\
        w_{ij}  &\leftarrow& \text{clip}(w_{ij},b^\text{min}_{ij},b^\text{max}_{ij})
        \end{eqnarray*}

    in case of additive noise.  Optionally, multiplicative noise can
    be chosen in which case the first equation becomes:

    .. math::

        w_{ij}  \leftarrow  w_{ij} - \Delta w_{ij}^d (\gamma_{ij}^d \;w_{ij} + 1)
        (1 + \sigma_\text{c-to-c}\,\xi)

    The cross-point `ij` dependent slope parameter
    :math:`\gamma_{ij}^d` are given during initialization by

    .. math::
        :nowrap:

        \begin{eqnarray*}
        \gamma_{ij}^+ &=& - |\gamma^+ + \gamma_\text{d-to-d}^+ \xi|/b^\text{max}_{ij}\\
        \gamma_{ij}^- &=& - |\gamma^- + \gamma_\text{d-to-d}^- \xi|/b^\text{min}_{ij}
        \end{eqnarray*}

    where the :math:`\xi` are standard Gaussian random variables and
    :math:`b^\text{min}_{ij}` and :math:`b^\text{max}_{ij}` the
    cross-point `ij` specific minimal and maximal weight bounds,
    respectively (see description for :class:`~PulsedDevice`).

    Note:
        If :math:`\gamma=1` and :math:`\gamma_\text{d-to-d}=0` this
        update implements `soft bounds`, since the updates step becomes
        equal to :math:`1/b`.

    Note:
        If :math:`\gamma=0` and :math:`\gamma_\text{d-to-d}=0` and
        additive noise, this update is identical to those described in
        :class:`~PulsedDevice`.
    """

    bindings_class: ClassVar[Type] = devices.LinearStepResistiveDeviceParameter

    gamma_up: float = 0.0
    r"""The value of :math:`\gamma^+`.

    Intuitively, a value of 0.1 means that the update step size in up direction
    at the weight bounds is 10% decreased relative to that origin :math:`w=0`.

    Note:
        In principle one could fix :math:`\gamma=\gamma^-=\gamma^+` since
        up/down variation can be given by ``up_down_dtod``, see
        :class:`~PulsedDevice`.

    Note:
        The hard-bounds are still observed, so that the weight cannot
        grow beyond its bounds.
    """

    gamma_down: float = 0.0
    r"""The value of :math:`\gamma^-`."""

    gamma_up_dtod: float = 0.05
    r"""Device-to-device variation for :math:`\gamma^+`, i.e. the value of
    :math:`\gamma_\text{d-to-d}^+`."""

    gamma_down_dtod: float = 0.05
    r"""Device-to-device variation for :math:`\gamma^-`, i.e. the value of
    :math:`\gamma_\text{d-to-d}^-`."""

    allow_increasing: bool = False
    """Whether to allow increasing of update sizes.

    Whether to allow the situation where update sizes increase towards the
    bound instead of saturating (and thus becoming smaller).
    """

    mean_bound_reference: bool = True
    r"""Whether to use instead of the above:

    .. math::

        \gamma_{ij}^+ &=& - |\gamma^+ + \gamma_\text{d-to-d}^+ \xi|/b^\text{max}

        \gamma_{ij}^- &=& - |\gamma^- + \gamma_\text{d-to-d}^- \xi|/b^\text{min}

    where :math:`b^\text{max}` and :math:`b^\text{max}` are the values given by
    ``w_max`` and ``w_min``, see :class:`~PulsedDevice`.
    """

    mult_noise: bool = True
    """Whether to use multiplicative noise instead of additive cycle-to-cycle
    noise."""

    write_noise_std: float = 0.0
    r"""Whether to use update write noise.

    Whether to use update write noise that is added to the updated
    devices weight, while the update is done on a hidden persistent weight. The
    update write noise is then sampled anew when the device is touched
    again.

    Thus it is:

    .. math::
        w_\text{apparent}{ij} = w_{ij} + \sigma_\text{write_noise} \Delta w_\text{min}\xi

    and the update is done on :math:`w_{ij}` but the forward sees the
    :math:`w_\text{apparent}`.
    """

    reverse_up: bool = False
    """Whether to increase the step size in up direction with increasing
    weights (default decreases).

    Note:
        If set, ``mult_noise`` needs to be also set.
    """

    reverse_down: bool = False
    """Whether to increase the step size in down direction with decreasing
    weights (default decreases).

    Note:
        If set, ``mult_noise`` needs to be also set.

    """

    reverse_offset: float = 0.01
    """Offset to add to the step size for reverse up or down to avoid
    zero step size at weight min or max.
    """


@dataclass
class SoftBoundsDevice(PulsedDevice):
    r"""Pulsed update behavioral model: soft bounds.

    Pulsed update behavioral model, where the update step response size
    of the material is linearly dependent and it goes to zero at the
    bound.

    This model is based on :class:`~LinearStepDevice` with
    parameters set to model soft bounds.
    """

    bindings_class: ClassVar[Type] = devices.SoftBoundsResistiveDeviceParameter

    mult_noise: bool = True
    """Whether to use multiplicative noise instead of additive cycle-to-cycle
    noise."""

    write_noise_std: float = 0.0
    r"""Whether to use update write noise.

    Whether to use update write noise that is added to the updated
    devices weight, while the update is done on a hidden persistent weight. The
    update write noise is then sampled anew when the device is touched
    again.

    Thus it is:

    .. math::
        w_\text{apparent}{ij} = w_{ij} + \sigma_\text{write_noise} \Delta w_\text{min}\xi

    and the update is done on :math:`w_{ij}` but the forward sees the
    :math:`w_\text{apparent}`.
    """

    reverse_up: bool = False
    """Whether to increase the step size in up direction with increasing
    weights (default decreases).

    Note:
        If set, ``mult_noise`` needs to be also set.
    """

    reverse_down: bool = False
    """Whether to increase the step size in down direction with decreasing
    weights (default decreases).

    Note:
        If set, ``mult_noise`` needs to be also set.
    """

    reverse_offset: float = 0.01
    """Offset to add to the step size for reverse up or down to avoid
    zero step size at weight min or max.
    """


@dataclass
class SoftBoundsPmaxDevice(SoftBoundsDevice):
    r"""Pulsed update behavioral model: soft bounds, with a different
    parameterization for easier device fitting to experimental data.

    Under the hood, the same  device behavior as :class:`~SoftboundsDevice`
    This model is based on :class:`~LinearStepDevice` with
    parameters set to model soft bounds.

    It implements pulse response function of the form:

    .. math::

        w(p_\text{up}) = B\left(1 -e^{-\alpha p_\text{up}} \right) + r_\text{min}

        w(p_\text{down}) = - B\left(1 - e^{-\alpha (p_\text{max}
        - p_\text{down})}\right) + r_\text{max}

    where :math:`B=\frac{r_\text{max} -
    r_\text{min}}{1 - e^{-\alpha p_\text{max}}}`.

    Here :math:`p_\text{max}` is the number of pulses that were applied to get
    the device from the minimum conductance (minimum of range,
    :math:`r_\text{min}`) to the maximum (maximum of range,
    :math:`r_\text{max}`).

    Internally the following transformation is used to get the
    original parameter of :class:`SoftboundsDevice`::

        b_factor = (range_max - range_min)/(1 - exp(-p_max * alpha))
        w_min = range_min
        w_max = range_min + b_factor
        dw_min = b_factor * alpha
        up_down = 1 + 2 * range_min / b_factor

    Note:
        Device-to-device and cycle-to-cycle variation are defined as
        before (see :class:`SoftBoundsDevice`, see also
        :class:`PulsedDevice`). That is, for instance `dw_min_dtod`
        will effectively change the slope (in units of ``dw_min`` which
        is ``b_factor * alpha``, see above). Range offset fluctuations
        can be achieved by using ``w_min_dtod`` and ``w_max_dtod``
        which will vary ``w_min`` and ``w_max`` across devices,
        respectively.
    """

    p_max: int = 1000
    """Number of pulses to drive the synapse from ``range_min`` to ``range_max``."""

    alpha: float = 0.001/2
    r"""The slope of the soft bounds model :math:`dw \propto \alpha w` for both
    up and down direction."""

    range_min: float = -1.0
    """Setting of the weight when starting the :math:`P_max` up pulse
    experiment."""

    range_max: float = 1.0
    """Value of the weight for :math:`P_max` number of up pulses."""

    #  these values will be set from the above, so we hide it.
    w_min: float = field(default_factory=lambda: None, metadata={'hide_if': None})  # type: ignore
    w_max: float = field(default_factory=lambda: None, metadata={'hide_if': None})  # type: ignore
    dw_min: float = field(default_factory=lambda: None, metadata={'hide_if': None})  # type: ignore
    up_down: float = field(default_factory=lambda: None, metadata={'hide_if': None})  # type: ignore

    def as_bindings(self) -> devices.PulsedResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        params = SoftBoundsDevice()
        for key, value in self.__dict__.items():
            if key not in ['range_min', 'range_max', 'alpha', 'p_max']:
                setattr(params, key, value)

        b_factor = (self.range_max - self.range_min)/(1 - exp(-self.p_max * self.alpha))
        params.w_min = self.range_min
        params.w_max = self.range_min + b_factor
        params.dw_min = b_factor * self.alpha
        params.up_down = 1 + 2 * self.range_min / b_factor

        return parameters_to_bindings(params)


@dataclass
class SoftBoundsReferenceDevice(PulsedDevice):
    r"""Pulsed update behavioral model: soft bounds with reference device.

    Pulsed update behavioral model, where the update step response size
    of the material is linearly dependent.

    In particular, the update behavior is

    .. math::
        \delta W_+ &=& \alpha_{+}(1 - \frac{w}{\bmax}) (1 + \sigma \xi)
        \delta W_- &=& \alpha_{-}(1 - \frac{w}{\bmin}) (1 + \sigma \xi)

    Where the same device-to-device varition can be given as for the
    ``PulsedDevice``. In addition, a device-to-device variation can be
    directly given on the slope.

    Moreover, a fixed reference conductance can be subtracted from the
    resulting weight, which implemented a differential read of $W -
    R$.

    """

    bindings_class: ClassVar[Type] = devices.SoftBoundsReferenceResistiveDeviceParameter

    mult_noise: bool = False
    """Whether to use multiplicative noise instead of additive cycle-to-cycle
    noise."""

    write_noise_std: float = 0.0
    r"""Whether to use update write noise.

    Whether to use update write noise that is added to the updated
    devices weight, while the update is done on a hidden persistent weight. The
    update write noise is then sampled anew when the device is touched
    again.

    Thus it is:

    .. math::
        w_\text{apparent}{ij} = w_{ij} + \sigma_\text{write_noise} \Delta w_\text{min}\xi

    and the update is done on :math:`w_{ij}` but the forward sees the
    :math:`w_\text{apparent}`.
    """

    slope_up_dtod: float = 0.0
    r"""Device-to-device variation on the up-pulse slope.

    Note:

        Since the up slope is proportional to
        :math:`\propto\frac{1}{b_\text{max}}` the device-to-device variation
        of the weight max ``w_max_dtod`` will also introduce a slope
        variation. Turn that off, if the variation should be only on
        the slope directly.

    """

    slope_down_dtod: float = 0.0
    r"""Device-to-device variation on the down-pulse slope.

    Note:

        Since the up slope is proportional to
        :math:`\propto\frac{1}{b_\text{min}}` the device-to-device variation
        of the weight max ``w_min_dtod`` will also introduce a slope
        variation. Turn that off, if the variation should be only on
        the slope directly.
    """

    reference_mean: float = 0.0
    """Added to all devices of the reference $R$.
    """

    reference_std: float = 0.0
    """Normal distributed device-to-device variation added to the reference $R$.
    """

    subtract_symmetry_point: bool = False
    r"""Whether to add the computed symmetry point of the devices onto the
    reference $R$.

    The symmetry point is given by:

    .. math::
        w_* = \frac{alpha_{+} - \alpha_{-}}{\frac{\alpha_{+}}{b_\text{max}}
                    - \frac{\alpha_{-}}{b_\text{min}}}
    """


@dataclass
class ExpStepDevice(PulsedDevice):
    r"""Exponential update step or CMOS-like update behavior.

    This model is derived from ``PulsedDevice`` and uses all its
    parameters. ``ExpStepDevice`` only implements a new 'update once'
    functionality, where the minimal weight step change with weight is
    fitted by an exponential function as detailed below.

    .. math::

        w_{ij}  \leftarrow  w_{ij} -  \max(y_{ij},0)  \Delta w_{ij}^d
        (1 + \sigma_\text{c-to-c}\,\xi)

    and :math:`y_{ij}` is given as

    .. math::
        z_{ij} = 2 a_\text{es} \frac{w_{ij}}{b^\text{max}_{ij} - b^\text{min}_{ij}}
        + b_\text{es}

        y_{ij} = 1 - A^{(d)} e^{d \gamma^{(d)} z_{ij}}

    where :math:`d` is the direction of the update (+ or -), see also
    :class:`~PulsedDevice` for details.

    All additional parameter (:math:`a_\text{es}`,
    :math:`b_\text{es}`, :math:`\gamma^{(d)}`, :math:`A^{(d)}` ) are
    tile-wise fitting parameters (ie. no device-to-device variation in
    these parameters).  Note that the other parameter involved can be
    still defined with device-to-device variation and (additional)
    up-down bias (see :class:`~PulsedDevice`).

    Note:
        This device also features a more complex cycle-to-cycle noise
        model of the update step, when specifying ``dw_min_std_add``
        and ``dw_min_std_slope``. By default,    The Gaussian noise added to the
        calculated update step size :math:`\Delta q_\text{act}` is
        proportional to

        .. math::
           \sigma_{\text{final}} = \sigma \left( \sigma_\text{add} +
           |\Delta w_\text{actual}| + \sigma_\text{slope} |w_\text{current}|\right)

        where the :math:`\sigma` is given by ``dw_min_std``,
        :math:`\sigma_\text{add}` is given by ``dw_min_std_add``, and
        :math:`\sigma_\text{slope}` is given by ``dw_min_std_slope``.
    """
    # pylint: disable=invalid-name

    bindings_class: ClassVar[Type] = devices.ExpStepResistiveDeviceParameter

    A_up: float = 0.00081
    """Factor ``A`` for the up direction."""

    A_down: float = 0.36833
    """Factor ``A`` for the down direction."""

    gamma_up: float = 12.44625
    """Exponent for the up direction."""

    gamma_down: float = 12.78785
    """Exponent for the down direction."""

    a: float = 0.244
    """Global slope parameter."""

    b: float = 0.2425
    """Global offset parameter."""

    dw_min_std_add: float = 0.0
    """additive cycle-to-cycle noise of the update size (in units of
    ``dw_min_std``, see above)."""

    dw_min_std_slope: float = 0.0
    """ cycle-to-cycle noise of the update size (in units of ``dw_min_std``, see above)."""

    write_noise_std: float = 0.0
    r"""Whether to use update write noise.

    Whether to use update write noise that is added to the updated
    devices weight, while the update is done on a hidden persistent weight. The
    update write noise is then sampled a new when the device is touched
    again.

    Thus it is:

    .. math::
        w_\text{apparent}{ij} = w_{ij} + \sigma_\text{write_noise}\xi

    and the update is done on :math:`w_{ij}` but the forward sees the
    :math:`w_\text{apparent}`.
    """


@dataclass
class PowStepDevice(PulsedDevice):
    r"""Pulsed update behavioral model: power-dependent step.

    Pulsed update behavioral model, where the update step response
    size of the material has a power-dependent with resistance. This
    device model implements (a shifted from of) the `Fusi & Abott
    (2007)`_ synapse model (see also `Frascaroli et al. (2108)`_).

    The model based on :class:`~PulsedDevice` and thus shares most
    parameters and functionality. However, it implements new `update
    once` function, where the update step size depends in the
    following way. If we set :math:`\omega_{ij} =
    \frac{b_{ij}^\text{max} - w_{ij}}{b_{ij}^\text{max} -
    b_{ij}^\text{min}}` the relative distance of the current weight to
    the upper bound, then the update per pulse is for the upwards direction:

    .. math::
        w_{ij}  \leftarrow  w_{ij} + \Delta w_{ij}^+\,(\omega_{ij})^{\gamma_{ij}^+}
        \left(1 + \sigma_\text{c-to-c}\,\xi\right)

    and in downwards direction:

    .. math::
        w_{ij}  \leftarrow  w_{ij} + \Delta w_{ij}^-\,(1 - \omega_{ij})^{\gamma_{ij}^-}
        \left(1 + \sigma_\text{c-to-c}\,\xi\right)

    Similar to :math:`\Delta w_{ij}^d` the exponent :math:`\gamma_{ij}` can be
    defined with device-to-device variation and bias in up and down
    direction:

    .. math::

        \gamma_{ij}^d = \gamma\, \left(1 + d\, \beta_{ij}
        + \sigma_\text{pow-gamma-d-to-d}\xi\right)

    where :math:`\xi` is again a standard Gaussian. :math:`\beta_{ij}`
    is the directional up `versus` down bias.  At initialization
    ``pow_up_down_dtod`` and ``pow_up_down`` defines this bias term:

    .. math::

        \beta_{ij} = \beta_\text{pow-up-down} + \xi\sigma_\text{pow-up-down-dtod}

    where :math:`\xi` is again a standard Gaussian number and
    :math:`\beta_\text{pow-up-down}` corresponds to ``pow_up_down``.

    Note:
        The ``pow_gamma_dtod`` and ``pow_up_down_dtod``
        device-to-device variation parameters are given in relative
        units to ``pow_gamma``.

    Note:
        :math:`\Delta w_{ij}^d` is defined as for the
        :class:`~PulsedDevice`, however, for this device, the update step
        size will *not* be given by :math:`\Delta w_{ij}` at
        :math:`w_{ij}=0` as for most other devices models

    ..  _Fusi & Abott (2007): https://www.nature.com/articles/nn1859
    ..  _Frascaroli et al. (2108): https://www.nature.com/articles/s41598-018-25376-x
    """

    bindings_class: ClassVar[Type] = devices.PowStepResistiveDeviceParameter

    pow_gamma: float = 1.0
    r"""The value of :math:`\gamma` as explained above.

    Note:
        :math:`\gamma` reduces essentially to the
        :class:`SoftBoundsDevice` (if no device-to-device variation of
        gamma is used additionally). However, the
        :class:`SoftBoundsDevice` will be much faster, as it does not
        need to compute the slow `pow` function.
    """

    pow_gamma_dtod: float = 0.1
    r"""Device-to-device variation for ``pow_gamma``.

    i.e. the value of :math:`\gamma_\text{pow-gamma-d-to-d}` given in relative
    units to ``pow_gamma``.
    """

    pow_up_down: float = 0.0
    r"""The up versus down bias of the :math:`\gamma` as described above.

    It is :math:`\gamma^+ = \gamma (1 + \beta_\text{pow-up-down})` and
    :math:`\gamma^- = \gamma (1 - \beta_\text{pow-up-down})` .
    """

    pow_up_down_dtod: float = 0.0
    r"""Device-to-device variation in the up versus down bias of
    :math:`\gamma` as descibed above.

    In units of ``pow_gamma``.
    """

    write_noise_std: float = 0.0
    r"""Whether to use update write noise.

    Whether to use update write noise that is added to the updated
    devices weight, while the update is done on a hidden persistent weight. The
    update write noise is then sampled a new when the device is touched
    again.

    Thus it is:

    .. math::
        w_\text{apparent}{ij} = w_{ij} + \sigma_\text{write_noise}\xi

    and the update is done on :math:`w_{ij}` but the forward sees the
    :math:`w_\text{apparent}`.
    """


@dataclass
class PowStepReferenceDevice(PulsedDevice):
    r"""Pulsed update behavioral model: power-dependent step.

    Pulsed update behavioral model, where the update step response
    size of the material has a power-dependent with resistance. This
    device model implements (a shifted from of) the `Fusi & Abott
    (2007)`_ synapse model (see also `Frascaroli et al. (2108)`_).

    This device model is identical to
    :class:`~aihwkit.simulator.configs.devices.PowStepDevice`,
    however, here it does not implement any write noise functionality
    but instead has an additional option to subtract a reference
    conductance from the weights which is statically set during
    initialization.

    The model based on :class:`~PulsedDevice` and thus shares most
    parameters and functionality. However, it implements new ``update
    once`` function, where the update step size depends in the
    following way. If we set :math:`\omega_{ij} =
    \frac{b_{ij}^\text{max} - w_{ij}}{b_{ij}^\text{max} -
    b_{ij}^\text{min}}` the relative distance of the current weight to
    the upper bound, then the update per pulse is for the upwards
    direction:

    .. math::
        w_{ij}  \leftarrow  w_{ij} + \Delta w_{ij}^+\,(\omega_{ij})^{\gamma_{ij}^+}
        \left(1 + \sigma_\text{c-to-c}\,\xi\right)

    and in downwards direction:

    .. math::
        w_{ij}  \leftarrow  w_{ij} + \Delta w_{ij}^-\,(1 - \omega_{ij})^{\gamma_{ij}^-}
        \left(1 + \sigma_\text{c-to-c}\,\xi\right)

    Similar to :math:`\Delta w_{ij}^d` the exponent :math:`\gamma_{ij}` can be
    defined with device-to-device variation and bias in up and down
    direction:

    .. math::

        \gamma_{ij}^d = \gamma\, \left(1 + d\, \beta_{ij}
        + \sigma_\text{pow-gamma-d-to-d}\xi\right)

    where :math:`\xi` is again a standard Gaussian. :math:`\beta_{ij}`
    is the directional up `versus` down bias.  At initialization
    ``pow_up_down_dtod`` and ``pow_up_down`` defines this bias term:

    .. math::

        \beta_{ij} = \beta_\text{pow-up-down} + \xi\sigma_\text{pow-up-down-dtod}

    where :math:`\xi` is again a standard Gaussian number and
    :math:`\beta_\text{pow-up-down}` corresponds to ``pow_up_down``.

    Note:
        The ``pow_gamma_dtod`` and ``pow_up_down_dtod``
        device-to-device variation parameters are given in relative
        units to ``pow_gamma``.

    Note:
        :math:`\Delta w_{ij}^d` is defined as for the
        :class:`~PulsedDevice`, however, for this device, the update step
        size will *not* be given by :math:`\Delta w_{ij}` at
        :math:`w_{ij}=0` as for most other devices models

    Note:

        In contrast to
        :class:`~aihwkit.simulator.configs.devices.PowStepDevice`,
        write noise is not supported.

    ..  _Fusi & Abott (2007): https://www.nature.com/articles/nn1859
    ..  _Frascaroli et al. (2108): https://www.nature.com/articles/s41598-018-25376-x

    """

    bindings_class: ClassVar[Type] = devices.PowStepReferenceResistiveDeviceParameter

    pow_gamma: float = 1.0
    r"""The value of :math:`\gamma` as explained above.

    Note:
        :math:`\gamma` reduces essentially to the
        :class:`SoftBoundsDevice` (if no device-to-device variation of
        gamma is used additionally). However, the
        :class:`SoftBoundsDevice` will be much faster, as it does not
        need to compute the slow `pow` function.
    """

    pow_gamma_dtod: float = 0.1
    r"""Device-to-device variation for ``pow_gamma``.

    i.e. the value of :math:`\gamma_\text{pow-gamma-d-to-d}` given in relative
    units to ``pow_gamma``.
    """

    pow_up_down: float = 0.0
    r"""The up versus down bias of the :math:`\gamma` as described above.

    It is :math:`\gamma^+ = \gamma (1 + \beta_\text{pow-up-down})` and
    :math:`\gamma^- = \gamma (1 - \beta_\text{pow-up-down})` .
    """

    pow_up_down_dtod: float = 0.0
    r"""Device-to-device variation in the up versus down bias of
    :math:`\gamma` as descibed above.

    In units of ``pow_gamma``.
    """

    subtract_symmetry_point: bool = False
    r"""Whether store the symmetry point of each device onto the reference device.

    The symmetry point is only numerically estimated, since an
    analytically for is not available due to the power step model.
    """

    n_estimation_steps: int = -1
    """The number of times to run an (noise-free) up / down pulse
    combination for the numerical estimation of the symmetry points.

    In case of a non-positive number, the number of estimations steps
    is set 10 times the expected number of states of each device
    (maxed out at 10000).
    """

    reference_mean: float = 0.0
    """Added to all devices of the reference :math:`r`.
    """

    reference_std: float = 0.0
    """Normal distributed device-to-device variation added to the reference :math:`r`.
    """


@dataclass
class PiecewiseStepDevice(PulsedDevice):
    r"""Piece-wise interpolated device update characteristics.

    This model is derived from :class:`~PulsedDevice` and uses all its
    parameters. :class:`~PiecewiseStepDevice` implements a new
    functionality where the device's update response curve is given
    explicitly on nodes over the weight range. The device will
    automatically interpolate the update step size using the given node
    values.

    In detail, the update in down direction of the device is given as:

    .. math::
        w_{ij} \leftarrow w_{ij} + \Delta w_{ij}^d \left((1 - q) v^d(i_w)
        + q \, v^d(i_w + 1)\right) (1 + \sigma_\text{c-to-c}\,\xi)

    where :math:`i_w` is the index of the given vector :math:`v^d`
    (``piecewise_down``) where the current weight value would fall
    into if scaled to the current min and max values of that device
    (first and last value are set at weight min and max, respectively,
    the other values are equally distributed in the range).

    The scalar :math:`q` is the relative position of the weight in the
    current segment between the two selected nodes :math:`v^d(i_w)`
    and :math:`v^d(i_w + 1)`.

    The update in up direction is computed analogously using the
    ``piecewise_up`` vector.

    Note:

        The piecewise up and down vectors need to have the same number
        of elements.

    Note:

        In case of GPUs the maximal number of nodes in the vectors
        is limited to below 64 due to performance reasons.

    """

    bindings_class: ClassVar[Type] = devices.PiecewiseStepResistiveDeviceParameter

    piecewise_up: List[float] = field(default_factory=lambda: [1])
    r"""Array of values that characterize the update steps in upwards direction.

    The values are equally spaced in ``w_min`` and `w_max`` (which
    could vary from device-to-device), where the first and the last
    value is set at the boundary. The update will be computed by
    linear interpolation of the adjacent values, depending on where
    the weight is currently within the range.

    The values are given as relative numbers: the final update size
    will be computed by multiplying the value with the current
    ``dw_min`` of the device.

    E.g.  ``[1.5, 1, 1.5]`` and ``dw_min=0.1`` means that the update
    (in up direction) is ``dw_min`` around zero weight value and
    linearly increasing to ``1.5 * dw_min`` for larger or smaller
    weight values.

    """

    piecewise_down: List[float] = field(default_factory=lambda: [1])
    r"""Array of values that characterize the update steps in downwards direction.

    Analogous to ``piecewise_up`` but for the downwards direction.
    """

    write_noise_std: float = 0.0
    r"""Whether to use update write noise.

    Whether to use update write noise that is added to the updated
    devices weight, while the update is done on a hidden persistent weight. The
    update write noise is then sampled a new when the device is touched
    again.

    Thus it is:

    .. math::
        w_\text{apparent}{ij} = w_{ij} + \sigma_\text{write_noise}\xi

    and the update is done on :math:`w_{ij}` but the forward sees the
    :math:`w_\text{apparent}`.
    """



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for inference tiles."""

from typing import Optional, List

from parameterized import parameterized
from torch import ones
from torch import Tensor
from torch.nn.functional import mse_loss
from torch.optim import SGD
from torch.nn import Linear

from aihwkit.nn import AnalogLinear
from aihwkit.optim import AnalogSGD
from aihwkit.simulator.configs import FloatingPointRPUConfig
from aihwkit.simulator.configs.utils import (
    WeightNoiseType,
    WeightClipType,
    WeightModifierType,
    WeightModifierParameter,
    WeightRemapType,
)
from aihwkit.inference import PCMLikeNoiseModel

from .helpers.decorators import parametrize_over_tiles
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import Inference, InferenceCuda


@parametrize_over_tiles([
    Inference,
    InferenceCuda
])
class InferenceTileTest(ParametrizedTestCase):
    """Inference model tests."""

    def get_model_and_x(self):
        """Trains a simple model."""
        # Prepare the datasets (input and expected output).
        x = Tensor([[0.1, 0.2, 0.4, 0.3], [0.2, 0.1, 0.1, 0.3]])
        y = Tensor([[1.0, 0.5], [0.7, 0.3]])

        # Define a single-layer network, using a constant step device type.
        rpu_config = self.get_rpu_config()
        rpu_config.forward.out_res = -1.  # Turn off (output) ADC discretization.
        rpu_config.forward.w_noise_type = WeightNoiseType.ADDITIVE_CONSTANT
        rpu_config.forward.w_noise = 0.02
        rpu_config.noise_model = PCMLikeNoiseModel(g_max=25.0)

        model = AnalogLinear(4, 2, bias=True, rpu_config=rpu_config)

        # Move the model and tensors to cuda if it is available.
        if self.use_cuda:
            x = x.cuda()
            y = y.cuda()
            model.cuda()

        # Define an analog-aware optimizer, preparing it for using the layers.
        opt = AnalogSGD(model.parameters(), lr=0.1)
        opt.regroup_param_groups(model)

        for _ in range(100):
            opt.zero_grad()

            # Add the training Tensor to the model (input).
            pred = model(x)
            # Add the expected output Tensor.
            loss = mse_loss(pred, y)
            # Run training (backward propagation).
            loss.backward()

            opt.step()

        return model, x

    def test_against_fp(self):
        """Test whether FP is same as is_perfect inference tile."""
        # pylint: disable-msg=too-many-locals
        # Prepare the datasets (input and expected output).
        x = Tensor([[0.1, 0.2, 0.4, 0.3], [0.2, 0.1, 0.1, 0.3]])
        y = Tensor([[1.0, 0.5], [0.7, 0.3]])

        # Define a single-layer network, using a constant step device type.
        rpu_config = self.get_rpu_config()
        rpu_config.forward.is_perfect = True
        model_torch = Linear(4, 2, bias=True)
        model = AnalogLinear(4, 2, bias=True, rpu_config=rpu_config)
        model.set_weights(model_torch.weight, model_torch.bias)
        model_fp = AnalogLinear(4, 2, bias=True, rpu_config=FloatingPointRPUConfig())
        model_fp.set_weights(model_torch.weight, model_torch.bias)

        self.assertTensorAlmostEqual(model.get_weights()[0], model_torch.weight)
        self.assertTensorAlmostEqual(model.get_weights()[0], model_fp.get_weights()[0])

        # Move the model and tensors to cuda if it is available.
        if self.use_cuda:
            x = x.cuda()
            y = y.cuda()
            model.cuda()
            model_fp.cuda()
            model_torch.cuda()

        # Define an analog-aware optimizer, preparing it for using the layers.
        opt = AnalogSGD(model.parameters(), lr=0.1)
        opt_fp = AnalogSGD(model_fp.parameters(), lr=0.1)
        opt_torch = SGD(model_torch.parameters(), lr=0.1)

        for _ in range(100):

            # inference
            opt.zero_grad()
            pred = model(x)
            loss = mse_loss(pred, y)
            loss.backward()
            opt.step()

            # same for fp
            opt_fp.zero_grad()
            pred_fp = model_fp(x)
            loss_fp = mse_loss(pred_fp, y)
            loss_fp.backward()
            opt_fp.step()

            # same for torch
            opt_torch.zero_grad()
            pred_torch = model_torch(x)
            loss_torch = mse_loss(pred_torch, y)
            loss_torch.backward()
            opt_torch.step()

            self.assertTensorAlmostEqual(pred_torch, pred)
            self.assertTensorAlmostEqual(loss_torch, loss)
            self.assertTensorAlmostEqual(model.get_weights()[0], model_torch.weight)

            self.assertTensorAlmostEqual(pred_fp, pred)
            self.assertTensorAlmostEqual(loss_fp, loss)
            self.assertTensorAlmostEqual(model.get_weights()[0], model_fp.get_weights()[0])

    def test_drift(self):
        """Test using realistic weights (bias)."""
        model, x = self.get_model_and_x()

        # do inference with drift
        pred_before = model(x)

        pred_last = pred_before
        model.eval()
        for t_inference in [0., 1., 20., 1000., 1e5]:
            model.drift_analog_weights(t_inference)
            pred_drift = model(x)
            self.assertNotAlmostEqualTensor(pred_last, pred_drift)
            pred_last = pred_drift

        self.assertNotAlmostEqualTensor(model.analog_tile.alpha, ones((1,)))

    def test_post_update_step_clip(self):
        """Tests whether post update diffusion is performed."""
        rpu_config = self.get_rpu_config()
        rpu_config.clip.type = WeightClipType.FIXED_VALUE
        rpu_config.clip.fixed_value = 0.3
        analog_tile = self.get_tile(2, 3, rpu_config=rpu_config, bias=True)

        weights = Tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
        biases = Tensor([-0.1, -0.6])

        analog_tile.set_learning_rate(0.123)
        analog_tile.set_weights(weights, biases)

        analog_tile.post_update_step()

        tile_weights, tile_biases = analog_tile.get_weights()

        self.assertNotAlmostEqualTensor(tile_weights, weights)
        self.assertNotAlmostEqualTensor(tile_biases, biases)

    def test_post_update_step_remap_layer(self):
        """Tests whether post update remap is performed."""
        rpu_config = self.get_rpu_config()
        rpu_config.mapping.out_scaling_columnwise = False
        rpu_config.mapping.learn_out_scaling = True
        rpu_config.mapping.weight_scaling_omega = 1.0
        rpu_config.mapping.weight_scaling_columnwise = False

        rpu_config.remap.type = WeightRemapType.LAYERWISE_SYMMETRIC
        analog_tile = self.get_tile(2, 3, rpu_config=rpu_config, bias=True)

        weights = Tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
        biases = Tensor([-0.1, -0.3])

        analog_tile.set_learning_rate(0.123)
        analog_tile.set_weights(weights, biases)

        analog_tile.post_update_step()

        tile_weights, tile_biases = analog_tile.get_weights()

        self.assertTensorAlmostEqual(tile_weights, weights / 0.6)
        self.assertTensorAlmostEqual(tile_biases, biases / 0.6)

        scales = analog_tile.get_mapping_scales()
        self.assertTensorAlmostEqual(scales, Tensor([0.6, 0.6]))

    def test_post_update_step_remap_column(self):
        """Tests whether post update remap is performed."""
        rpu_config = self.get_rpu_config()
        rpu_config.mapping.out_scaling_columnwise = True
        rpu_config.mapping.learn_out_scaling = True
        rpu_config.mapping.weight_scaling_omega = 1.0
        rpu_config.mapping.weight_scaling_columnwise = True

        rpu_config.remap.type = WeightRemapType.CHANNELWISE_SYMMETRIC
        analog_tile = self.get_tile(2, 3, rpu_config=rpu_config, bias=True)

        weights = Tensor([[-0.7, 0.2, 0.3], [0.4, 0.5, 0.6]])
        biases = Tensor([-0.1, -0.3])

        analog_tile.set_learning_rate(0.123)
        analog_tile.set_weights(weights, biases)

        analog_tile.post_update_step()

        tile_weights, tile_biases = analog_tile.get_weights()

        expected_scales = Tensor([0.7, 0.6])
        self.assertTensorAlmostEqual(tile_weights, weights / expected_scales.view(-1, 1))
        self.assertTensorAlmostEqual(tile_biases, biases / expected_scales)

        scales = analog_tile.get_mapping_scales()
        self.assertTensorAlmostEqual(scales, expected_scales)

    @parameterized.expand([
        ('none', None,),
        ('dorefa', WeightModifierType.DOREFA,),
        ('mult_normal', WeightModifierType.MULT_NORMAL,),
        ('poly', WeightModifierType.POLY, [1., 3.]),
        ('polyN', WeightModifierType.POLY, [0.1, 0.2, 0.2, 0.3]),
        ('discretize', WeightModifierType.DISCRETIZE,),
        ('add_normal', WeightModifierType.DISCRETIZE_ADD_NORMAL,),
    ])
    def test_post_forward_modifier_types(self, _,
                                         modifier_type: 'WeightModifierType',
                                         coeffs: Optional[List] = None):
        """Tests whether modifier is performed."""
        rpu_config = self.get_rpu_config()
        rpu_config.drift_compensation = None
        rpu_config.forward.is_perfect = True
        rpu_config.forward.out_noise = 0.0
        rpu_config.forward.inp_noise = 0.0

        modifier = self.get_modifier(modifier_type, coeffs)
        if modifier is not None:
            rpu_config.modifier = modifier

        analog_tile = self.get_tile(2, 3, rpu_config=rpu_config, bias=True)
        x_input = Tensor([[0.1, 0.2, 0.3]])
        weights = Tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
        biases = Tensor([-0.1, -0.2])

        if self.use_cuda:
            x_input = x_input.cuda()

        analog_tile.set_learning_rate(0.123)
        analog_tile.set_weights(weights, biases)

        x_output = analog_tile.forward(x_input, is_test=True)
        analog_tile.post_update_step()
        x_output_post = analog_tile.forward(x_input, is_test=False)
        x_output_post_true = analog_tile.forward(x_input, is_test=True)
        tile_weights, tile_biases = analog_tile.get_weights()

        self.assertTensorAlmostEqual(tile_weights, weights)
        self.assertTensorAlmostEqual(tile_biases, biases)

        if modifier is None:
            self.assertTensorAlmostEqual(x_output, x_output_post)
        else:
            self.assertNotAlmostEqualTensor(x_output, x_output_post)

        self.assertTensorAlmostEqual(x_output, x_output_post_true)

    @staticmethod
    def get_modifier(
            modifier_type: Optional[WeightModifierType],
            coeffs: Optional[List] = None,
    ) -> Optional[WeightModifierParameter]:
        """Returns the modifier parameter."""
        if modifier_type is None:
            return None
        if coeffs is None:
            coeffs = [1.0, 0.1]

        modifier = WeightModifierParameter(
            type=modifier_type,
            std_dev=1.0,
            enable_during_test=False,
            res=0.132,
            coeffs=coeffs,
            rel_to_actual_wmax=False,
            assumed_wmax=1.0
        )

        if modifier_type == WeightModifierType.COPY:
            modifier.pdrop = 0.5

        return modifier



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Session handler for the AIHW Composer API."""

from typing import Any, Text, Union, TYPE_CHECKING

from requests import HTTPError, Session
import urllib3
from urllib3.exceptions import InsecureRequestWarning

from aihwkit.version import __version__
from aihwkit.cloud.client.exceptions import ApiResponseError, ResponseError

if TYPE_CHECKING:
    from typing import Optional


class ObjectStorageSession(Session):
    """Session handler for requests to object storage."""

    def request(  # type: ignore
            self,
            method: str,
            url: Union[str, bytes, Text],
            *args: Any,
            **kwargs: Any
    ) -> Any:
        """Construct a Request, prepares it and sends it.

        Args:
            method: method for the new ``Request`` object.
            url: URL for the new ``Request`` object.
            args: additional arguments for the original ``requests`` method.
            kwargs: additional arguments for the original ``requests`` method.

        Returns:
            A new ``Response`` object.

        Raises:
            ResponseError: if the response did not have a valid status code.
        """
        # pylint: disable=signature-differs
        response = super().request(method, url, *args, **kwargs)

        try:
            response.raise_for_status()
        except HTTPError:
            raise ResponseError(response) from None

        return response


class ApiSession(Session):
    """Session handler for requests to the AIHW Composer API.

    Custom ``Session`` for interfacing with the AIHW Composer API, using:

    * authorization based on jwt token.
    * custom user agent for the requests.

    Additionally, this class stores information about the API URL and base
    token.
    """

    def __init__(
            self,
            api_url: str,
            api_token: str,
            verify: bool = True
    ):
        super().__init__()

        self.api_url = api_url
        self.api_token = api_token
        self.verify = verify
        if not verify:
            urllib3.disable_warnings(InsecureRequestWarning)  # type: ignore[no-untyped-call]

        self.jwt_token = None  # type: Optional[str]

        self.headers.update({'User-Agent': 'aihwkit/{}'.format(__version__)})

    def update_jwt_token(self, jwt_token: str) -> None:
        """Set the jwt token for the session."""
        self.jwt_token = jwt_token
        self.headers.update({'Authorization': 'Bearer {}'.format(jwt_token)})

    def request(  # type: ignore
            self,
            method: str,
            url: Union[str, bytes, Text],
            *args: Any,
            **kwargs: Any
    ) -> Any:
        """Construct a Request, prepares it and sends it.

        Args:
            method: method for the new ``Request`` object.
            url: URL for the new ``Request`` object.
            args: additional arguments for the original ``requests`` method.
            kwargs: additional arguments for the original ``requests`` method.

        Returns:
            A new ``Response`` object.

        Raises:
            ApiResponseError: if the response did not have a valid status code.
        """
        # pylint: disable=signature-differs
        full_url = '{}/{}'.format(self.api_url, str(url))

        response = super().request(method, full_url, *args, **kwargs)

        try:
            response.raise_for_status()
        except HTTPError:
            raise ApiResponseError(response) from None

        return response



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for RNN layers."""
from torch import randn, ones, no_grad
from torch.nn import MSELoss
from numpy.testing import assert_array_almost_equal, assert_raises

from aihwkit.optim import AnalogSGD
from aihwkit.optim.context import AnalogContext

from .helpers.decorators import parametrize_over_layers
from .helpers.layers import LSTM, LSTMCuda, GRU, GRUCuda, VanillaRNN, VanillaRNNCuda, \
                            LSTMCombinedWeight, LSTMCombinedWeightCuda
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import FloatingPoint, Inference


@parametrize_over_layers(
    layers=[LSTM, VanillaRNN, GRU, LSTMCuda, GRUCuda, VanillaRNNCuda],
    tiles=[FloatingPoint, Inference],
    biases=['analog', 'digital', None]
)
class RNNLayerTest(ParametrizedTestCase):
    """ Base test for RNNs"""

    @staticmethod
    def train_once(model, y_in, y_out, analog_if, lr=0.5, digital_bias_lr_scale=1.0):
        """Train once."""

        criterion = MSELoss()
        optimizer = AnalogSGD(model.parameters(), lr=lr, momentum=0.0, nesterov=0.0)
        optimizer.regroup_param_groups()
        batch_size = y_in.size()[1]

        if analog_if:
            for param_group in optimizer.param_groups:
                if isinstance(param_group['params'][0], AnalogContext):
                    param_group['lr'] = lr
                else:
                    param_group['lr'] = lr * digital_bias_lr_scale
            states = model.get_zero_state(batch_size)

        else:
            states = None

        for _ in range(2):
            optimizer.zero_grad()
            pred, _ = model(y_in, states)
            loss = criterion(pred.mean(axis=2, keepdim=True), y_out)
            loss.backward()
            optimizer.step()

        return pred.detach().cpu().numpy()

    @staticmethod
    def train_once_bidir(model, y_in, y_out, analog_if):
        """Train once."""
        criterion = MSELoss()
        optimizer = AnalogSGD(model.parameters(), lr=0.5, momentum=0.0, nesterov=0.0)
        batch_size = y_in.size()[1]

        if analog_if:
            states = model.get_zero_state(batch_size)
        else:
            states = None

        for _ in range(2):
            optimizer.zero_grad()
            pred, _ = model(y_in, states)
            loss = criterion(pred.mean(axis=2, keepdim=True), y_out)
            loss.backward()
            optimizer.step()

        return pred.detach().cpu().numpy()

    def test_layer_instantiation(self):
        """Test AnalogLSTM layer instantiation."""
        input_size = 2
        hidden_size = 3
        num_layers = 4

        model = self.get_layer(input_size=input_size,
                               hidden_size=hidden_size,
                               num_layers=num_layers)

        # Assert over the stacked layers.
        self.assertEqual(len(model.rnn.layers), num_layers)
        for i, layer in enumerate(model.rnn.layers):
            # Assert over the size of weight_ih.
            if i == 0:
                self.assertEqual(layer.cell.weight_ih.in_features, input_size)
            else:
                self.assertEqual(layer.cell.weight_ih.in_features, hidden_size)
            self.assertEqual(layer.cell.weight_hh.in_features, hidden_size)
            # Assert over the rpu_config.
            analog_tile_ih = list(layer.cell.weight_ih.analog_tiles())[0]
            analog_tile_hh = list(layer.cell.weight_hh.analog_tiles())[0]
            self.assertEqual(analog_tile_ih.rpu_config.__class__,
                             self.get_rpu_config().__class__)
            self.assertEqual(analog_tile_hh.rpu_config.__class__,
                             self.get_rpu_config().__class__)

    def get_native_layer_comparison(self, *args, **kwargs):
        """ Returns the torch native model """
        raise NotImplementedError

    def test_layer_training(self):
        """Test AnalogLSTM layer training."""
        # pylint: disable=too-many-locals, too-many-statements
        def get_parameters(model, analog_if) -> dict:
            """Returns the parameter in an dict."""

            dic = {}
            for name, param in model.named_parameters():
                if isinstance(param, AnalogContext):
                    weight, bias = param.analog_tile.get_weights()
                    splits = name.split('.')
                    add_on = '_' + splits[-2].split('_')[-1] + '_l' + splits[2]
                    dic['weight' + add_on] = weight
                    if bias is not None:
                        dic['bias' + add_on] = bias
                elif analog_if and name.endswith('bias'):  # digital bias
                    splits = name.split('.')
                    add_on = '_' + splits[-2].split('_')[-1] + '_l' + splits[2]
                    dic['bias' + add_on] = param
                else:
                    dic[name] = param

            return dic

        input_size = 2
        hidden_size = 2
        num_layers = 2
        seq_length = 3
        batch_size = 3
        test_for_update = False  # For debugging. Does test whether all weights are updated.

        # Make dataset (just random).
        y_in = randn(seq_length, batch_size, input_size)
        y_out = ones(seq_length, batch_size, 1)

        rnn_analog = self.get_layer(input_size=input_size,
                                    hidden_size=hidden_size,
                                    num_layers=num_layers,
                                    realistic_read_write=False,
                                    dropout=0.0)

        rnn = self.get_native_layer_comparison(input_size=input_size,
                                               hidden_size=hidden_size,
                                               num_layers=num_layers,
                                               dropout=0.0,
                                               bias=self.bias)

        weights_org = []

        # pylint: disable=protected-access
        rnn_analog._apply_to_analog(lambda lay: weights_org.append(
            lay.analog_tile.tile.get_weights()))

        rnn_pars0 = get_parameters(rnn, False)
        rnn_analog_pars0 = get_parameters(rnn_analog, True)

        for par_name, par_item in rnn_pars0.items():
            par_item.data = rnn_analog_pars0[par_name].detach().clone()

        # Make independent for comparison below.
        rnn_pars0 = {key: value.detach().clone() for key, value in rnn_pars0.items()}

        if self.use_cuda:
            y_in = y_in.cuda()
            y_out = y_out.cuda()
            rnn_analog.cuda()
            rnn.cuda()

        with no_grad():
            assert_array_almost_equal(rnn(y_in)[0].detach().clone().cpu(),
                                      rnn_analog(y_in)[0].detach().clone().cpu())

        # First train analog and make sure weights differ.
        pred_analog = self.train_once(rnn_analog, y_in, y_out, True)

        analog_weights = []
        rnn_analog._apply_to_analog(lambda lay: analog_weights.append(
            lay.analog_tile.tile.get_weights()))

        if test_for_update:
            for weight, weight_org in zip(analog_weights, weights_org):
                assert_raises(AssertionError, assert_array_almost_equal, weight, weight_org)

        # Compare with RNN.
        pred = self.train_once(rnn, y_in, y_out, False)
        assert_array_almost_equal(pred, pred_analog)

        rnn_analog._apply_to_analog(lambda lay: lay._sync_weights_from_tile())

        rnn_pars = get_parameters(rnn, False)
        rnn_analog_pars = get_parameters(rnn_analog, True)

        if test_for_update:
            for par_name, par_item in rnn_pars.items():
                par0 = rnn_pars0[par_name].detach().cpu().numpy()
                par = par_item.detach().cpu().numpy()
                assert_raises(AssertionError, assert_array_almost_equal, par, par0)

        for par_name, par_item in rnn_pars.items():
            assert_array_almost_equal(par_item.detach().cpu().numpy(),
                                      rnn_analog_pars[par_name].detach().cpu().numpy())

    def test_bidir_layer_training(self):
        """Test AnalogLSTM bidirectional layer training."""
        # pylint: disable=too-many-locals, too-many-statements
        def get_parameters(model, analog_if) -> dict:
            """Returns the parameter in an dict."""
            dic = {}
            for name, param in model.named_parameters():
                if isinstance(param, AnalogContext):
                    weight, bias = param.analog_tile.get_weights()
                    splits = name.split('.')
                    add_on = '_' + splits[-2].split('_')[-1] + '_l' + splits[2]
                    if splits[4] == '1':
                        add_on += '_reverse'

                    dic['weight' + add_on] = weight
                    if bias is not None:
                        dic['bias' + add_on] = bias
                elif analog_if and name.endswith('bias'):  # digital bias
                    splits = name.split('.')
                    add_on = '_' + splits[-2].split('_')[-1] + '_l' + splits[2]
                    if splits[4] == '1':
                        add_on += '_reverse'
                    dic['bias' + add_on] = param
                else:
                    dic[name] = param

            return dic

        input_size = 4
        hidden_size = 5
        num_layers = 2
        seq_length = 10
        batch_size = 3
        test_for_update = False  # For debugging. Does test whether all weights are updated.

        # Make dataset (just random).
        y_in = randn(seq_length, batch_size, input_size)
        y_out = ones(seq_length, batch_size, 1)

        rnn_analog = self.get_layer(input_size=input_size,
                                    hidden_size=hidden_size,
                                    num_layers=num_layers,
                                    realistic_read_write=False,
                                    dropout=0.0,
                                    bidir=True)

        rnn = self.get_native_layer_comparison(input_size=input_size,
                                               hidden_size=hidden_size,
                                               num_layers=num_layers,
                                               dropout=0.0,
                                               bias=self.bias,
                                               bidirectional=True)

        weights_org = []

        # pylint: disable=protected-access
        rnn_analog._apply_to_analog(lambda lay: weights_org.append(
            lay.analog_tile.tile.get_weights()))

        rnn_pars0 = get_parameters(rnn, False)
        rnn_analog_pars0 = get_parameters(rnn_analog, True)

        for par_name, par_item in rnn_pars0.items():
            par_item.data = rnn_analog_pars0[par_name].detach().clone()

        # Make independent for comparison below.
        rnn_pars0 = {key: value.detach().clone() for key, value in rnn_pars0.items()}

        if self.use_cuda:
            y_in = y_in.cuda()
            y_out = y_out.cuda()
            rnn_analog.cuda()
            rnn.cuda()

        with no_grad():
            assert_array_almost_equal(rnn(y_in)[0].detach().clone().cpu(),
                                      rnn_analog(y_in)[0].detach().clone().cpu())

        # First train analog and make sure weights differ.
        pred_analog = self.train_once_bidir(rnn_analog, y_in, y_out, True)

        analog_weights = []
        rnn_analog._apply_to_analog(lambda lay: analog_weights.append(
            lay.analog_tile.tile.get_weights()))

        if test_for_update:
            for weight, weight_org in zip(analog_weights, weights_org):
                assert_raises(AssertionError, assert_array_almost_equal, weight, weight_org)

        # Compare with RNN.
        pred = self.train_once_bidir(rnn, y_in, y_out, False)
        assert_array_almost_equal(pred, pred_analog)

        rnn_analog._apply_to_analog(lambda lay: lay._sync_weights_from_tile())

        rnn_pars = get_parameters(rnn, False)
        rnn_analog_pars = get_parameters(rnn_analog, True)

        if test_for_update:
            for par_name, par_item in rnn_pars.items():
                par0 = rnn_pars0[par_name].detach().cpu().numpy()
                par = par_item.detach().cpu().numpy()
                assert_raises(AssertionError, assert_array_almost_equal, par, par0)

        for par_name, par_item in rnn_pars.items():
            assert_array_almost_equal(par_item.detach().cpu().numpy(),
                                      rnn_analog_pars[par_name].detach().cpu().numpy())


@parametrize_over_layers(
    layers=[LSTMCombinedWeight, LSTMCombinedWeightCuda],
    tiles=[FloatingPoint],
    biases=['digital', None]
)
class LSTMCombinedWeightTest(RNNLayerTest):
    """ Base test for AnalogLSTMCombinedWeight"""

    def test_layer_instantiation(self):
        """Test AnalogLSTM layer instantiation."""
        input_size = 2
        hidden_size = 3
        num_layers = 4

        model = self.get_layer(input_size=input_size,
                               hidden_size=hidden_size,
                               num_layers=num_layers)

        # Assert over the stacked layers.
        self.assertEqual(len(model.rnn.layers), num_layers)
        for i, layer in enumerate(model.rnn.layers):
            # Assert over the size of weight_ih.
            if i == 0:
                self.assertEqual(layer.cell.weight.in_features, input_size+hidden_size)
            else:
                self.assertEqual(layer.cell.weight.in_features, 2 * hidden_size)
            self.assertEqual(layer.cell.weight.out_features, 4 * hidden_size)
            # Assert over the rpu_config.
            analog_tile = list(layer.cell.weight.analog_tiles())[0]
            self.assertEqual(analog_tile.rpu_config.__class__,
                             self.get_rpu_config().__class__)

    def get_native_layer_comparison(self, *args, **kwargs):
        """ Returns the torch native model """
        raise NotImplementedError

    def test_layer_training(self):
        """Test AnalogLSTM layer training."""
        # pylint: disable=too-many-locals, too-many-statements
        def get_parameters(model, analog_if) -> dict:
            """Returns the parameter in an dict."""

            dic = {}
            for name, param in model.named_parameters():
                if isinstance(param, AnalogContext):
                    weight, bias = param.analog_tile.get_weights()
                    lay = int(name.split('.')[2])
                    add_on = '_l' + str(lay)
                    if lay == 0:
                        in_dim = input_size
                    else:
                        in_dim = hidden_size
                    dic['weight_ih' + add_on] = weight[..., :in_dim]
                    dic['weight_hh' + add_on] = weight[..., in_dim:]

                    if bias is not None:
                        dic['bias_ih' + add_on] = bias
                        dic['bias_hh' + add_on] = 0.0 * bias

                elif analog_if and name.endswith('bias'):  # digital bias
                    add_on = '_l' + name.split('.')[2]
                    dic['bias_ih' + add_on] = param
                    dic['bias_hh' + add_on] = 0.0 * param
                else:
                    dic[name] = param

            return dic

        input_size = 5
        hidden_size = 3
        num_layers = 2
        seq_length = 10
        batch_size = 4
        test_for_update = True  # For debugging. Does test whether all weights are updated.

        # Make dataset (just random).
        y_in = randn(seq_length, batch_size, input_size)
        y_out = ones(seq_length, batch_size, 1)

        rnn_analog = self.get_layer(input_size=input_size,
                                    hidden_size=hidden_size,
                                    num_layers=num_layers,
                                    realistic_read_write=False,
                                    dropout=0.0)

        rnn = self.get_native_layer_comparison(input_size=input_size,
                                               hidden_size=hidden_size,
                                               num_layers=num_layers,
                                               dropout=0.0,
                                               bias=self.bias)

        rnn_pars0 = get_parameters(rnn, False)
        rnn_analog_pars0 = get_parameters(rnn_analog, True)

        for par_name, par_item in rnn_pars0.items():
            par_item.data = rnn_analog_pars0[par_name].detach().clone()

        if test_for_update:
            weights_org = []
            # pylint: disable=protected-access
            rnn_analog._apply_to_analog(lambda lay: weights_org.append(
                lay.get_weights()))

        if self.use_cuda:
            y_in = y_in.cuda()
            y_out = y_out.cuda()
            rnn_analog.cuda()
            rnn.cuda()

        with no_grad():
            assert_array_almost_equal(rnn(y_in)[0].detach().clone().cpu(),
                                      rnn_analog(y_in)[0].detach().clone().cpu())

        # First train analog and make sure weights differ.
        # since there is only one bias the LR of the bias is changed
        pred_analog = self.train_once(rnn_analog, y_in, y_out, True, digital_bias_lr_scale=2.0)

        if test_for_update:
            analog_weights = []
            # pylint: disable=protected-access
            rnn_analog._apply_to_analog(lambda lay: analog_weights.append(
                lay.get_weights()))
            for weight, weight_org in zip(analog_weights, weights_org):
                assert_raises(AssertionError, assert_array_almost_equal, weight[0], weight_org[0])

        # Compare with RNN.
        pred = self.train_once(rnn, y_in, y_out, False)

        assert_array_almost_equal(pred, pred_analog)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Compound configurations presets for resistive processing units."""

# pylint: disable=too-many-instance-attributes
from typing import List
from dataclasses import dataclass, field

from aihwkit.simulator.configs.compounds import OneSidedUnitCell
from aihwkit.simulator.configs.utils import IOParameters, UpdateParameters
from aihwkit.simulator.presets.devices import PCMPresetDevice
from aihwkit.simulator.presets.utils import (
    PresetIOParameters, PresetUpdateParameters
)


@dataclass
class PCMPresetUnitCell(OneSidedUnitCell):
    """A unit cell that is comprised of two uni-directional PCM devices of
    opposite sign (see :class:`~PCMPresetDevice`).

    Check for refresh is performed after each mini-batch update. See
    :class:`~aihwkit.simulator.configs.device.OneSidedUnitCell` for
    details on the refresh implementation.
    """

    unit_cell_devices: List = field(
        default_factory=lambda: [PCMPresetDevice(), PCMPresetDevice()])

    refresh_every: int = 1
    units_in_mbatch: bool = True
    refresh_forward: IOParameters = field(default_factory=PresetIOParameters)
    refresh_update: UpdateParameters = field(
        default_factory=lambda: PresetUpdateParameters(desired_bl=31))



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utilities for inference."""

from torch.nn import Module
from aihwkit.exceptions import ModuleError


def drift_analog_weights(model: Module, t_inference: float = 0.0) -> None:
    """(Program) and drift all analog inference layers of a given model.

    Args:
        model: torch model with analog layers
        t_inference: assumed time of inference (in sec)

    Raises:
        ModuleError: if the layer is not in evaluation mode.
    """
    # avoid circular import
    # pylint: disable=import-outside-toplevel
    from aihwkit.nn.modules.base import AnalogModuleBase
    if model.training:
        raise ModuleError('drift_analog_weights can only be applied in '
                          'evaluation mode')

    for module in model.modules():
        if not isinstance(module, AnalogModuleBase):
            continue
        module.drift_analog_weights(t_inference)


def program_analog_weights(model: Module) -> None:
    """Program all analog inference layers of a given model.

    Args:
        model: torch model with analog layers

    Raises:
        ModuleError: if the layer is not in evaluation mode.
    """
    # avoid circular import
    # pylint: disable=import-outside-toplevel
    from aihwkit.nn.modules.base import AnalogModuleBase
    if model.training:
        raise ModuleError('program_analog_weights can only be applied in '
                          'evaluation mode')

    for module in model.modules():
        if not isinstance(module, AnalogModuleBase):
            continue
        module.program_analog_weights()
